# SPDX-License-Identifier: MIT

from neuralmag.common import config
from neuralmag.common import engine as en
from neuralmag.common.code_class import CodeClass

__all__ = ["Function", "VectorFunction", "CellFunction", "VectorCellFunction"]


class Function(CodeClass):
    """
    This class represents a discretized field on the mesh of a state object.

    If the instance is not intialized with a tensor, the tensor is lazy-
    initialized with zeros on the first access.

    :param state: The state that is used to construct the function
    :type state: :class:`State`
    :param spaces: The function spaces in the principal direction (defaults to
        "nnn" for 3D meshes and "nn" for 2D meshes)
    :type spaces: str
    :param shape: The shape of the function (either ``()`` for scalars or ``(3,)`` for
        vectors is currently supported)
    :type shape: tuple
    :param tensor: Tensor with discretized function values
    :type tensor: :class:`torch.Tensor`
    :param name: Name of the function
    :type name: str
    """

    def __init__(self, state, spaces=None, shape=(), tensor=None, name=None):
        self._state = state
        if spaces is None:
            spaces = "n" * state.mesh.dim
        self._spaces = spaces
        self._shape = shape
        if name is None:
            self._name = "f"
        else:
            self._name = name

        tensor_shape = []
        for i, space in enumerate(spaces):
            if space == "c":
                tensor_shape.append(state.mesh.n[i])
            elif space == "n":
                tensor_shape.append(state.mesh.n[i] + 1)
            else:
                raise Exception(f"Function space '{space}' not supported")

        self._tensor_shape = tuple(tensor_shape) + shape

        if tensor is None or isinstance(tensor, config.backend.Tensor):
            self._tensor = tensor
        else:
            raise NotImplemented("Unsupported tensor type.")
        self._expanded = False

        self.save_and_load_code(spaces, shape)

    @property
    def name(self):
        """
        The name of the function
        """
        return self._name

    @property
    def shape(self):
        """
        The shape of the function
        """
        return self._shape

    @property
    def tensor_shape(self):
        """
        The shape of the tensor with the discretized field values
        """
        return self._tensor_shape

    @property
    def state(self):
        """
        The state object used for the construction of the function
        """
        return self._state

    @property
    def spaces(self):
        """
        The function spaces of the function
        """
        return self._spaces

    @property
    def tensor(self):
        """
        The tensor containing the discretized values of the function
        """
        if self._tensor is None:
            self._tensor = config.backend.zeros(
                self._tensor_shape, dtype=self._state.dtype, device=self._state.device
            )
        return self._tensor

    @tensor.setter
    def tensor(self, tensor):
        self._tensor = tensor

    def fill(self, constant, expand=False):
        """
        Fills the tensor of the function with a constant value.

        :param constant: The constant to fill the tensor
        :type constant: int, list
        :param expand: If True, the tensor is set by expanding the constant
            to the size of the mesh using :code:`torch.Tensor.expand`
            resulting in minimal storage consumption.
        :type expand: bool
        :return: The function itself
        :rvalue: :class:`Function`

        :Example:
            .. code-block::

                state = nm.State(nm.Mesh((10, 10, 10), (1e-9, 1e-9, 1e-9)))
                f = nm.Function(state, shape = (3,)).fill([1.0, 2.0, 3.0])
        """
        tensor = self.state.tensor(constant)
        shape = self._tensor_shape
        if self.shape == (3,) and expand == False:
            shape = self._tensor_shape[:-1] + (1,)

        if expand:
            self._tensor = config.backend.broadcast_to(tensor, shape)
        else:
            self._tensor = config.backend.tile(tensor, shape)
        return self

    def avg(self):
        """
        Returns the componentwise average of the function over the mesh.

        :return: The componentwise average
        :rtype: :class:`torch.Tensor`
        """
        return self._code.avg(self._state.rho.tensor, self._state.dx, self.tensor)

    @classmethod
    def _generate_code(cls, spaces, shape):
        code = config.backend.CodeBlock()
        dim = len(spaces)

        # generate avg method
        f = en.Variable("f", spaces, shape)
        with code.add_function("avg", ["rho", "dx", "f"]) as func:
            terms, _ = en.compile_functional(1 * en.dV(dim))
            func.assign_sum("vol", *[term["cmd"] for term in terms])

            if shape == ():
                terms, variables = en.compile_functional(f * en.dV(dim))
                func.assign_sum("fint", *[term["cmd"] for term in terms])
            elif shape == (3,):
                func.zeros_like("fint", "f", (3,))
                for i in range(3):
                    terms, _ = en.compile_functional(f.dot(en.cs_e[i]) * en.dV(dim))
                    func.assign_sum(f"fint", *[term["cmd"] for term in terms], index=i)

            func.retrn("fint / vol")

        return code


class CellFunction(Function):
    """
    Subclass of :class:`Function` with the function space set to cellwise in each dimension.

    :param state: The state that is used to construct the function
    :type state: :class:`State`
    :type spaces: str
    :param shape: The shape of the function (either ``()`` for scalars or ``(3,)`` for
        vectors is currently supported)
    :type shape: tuple
    :param tensor: Tensor with discretized function values
    :type tensor: :class:`torch.Tensor`
    :param name: Name of the function
    :type name: str
    """

    def __init__(self, state, **kwargs):
        assert "spaces" not in kwargs
        kwargs["spaces"] = "c" * state.mesh.dim
        super().__init__(state, **kwargs)


class VectorFunction(Function):
    """
    Subclass of :class:`Function` with the shape set to (3,).

    :param state: The state that is used to construct the function
    :type state: :class:`State`
    :param spaces: The function spaces in the principal direction (defaults to
        "nnn" for 3D meshes and "nn" for 2D meshes)
    :type spaces: str
    :param tensor: Tensor with discretized function values
    :type tensor: :class:`torch.Tensor`
    :param name: Name of the function
    :type name: str
    """

    def __init__(self, state, **kwargs):
        assert "shape" not in kwargs
        kwargs["shape"] = (3,)
        super().__init__(state, **kwargs)


class VectorCellFunction(Function):
    """
    Subclass of :class:`Function` with the shape set to (3,) and the
    function space set to cellwise in each dimension.

    :param state: The state that is used to construct the function
    :type state: :class:`State`
    :param tensor: Tensor with discretized function values
    :type tensor: :class:`torch.Tensor`
    :param name: Name of the function
    :type name: str
    """

    def __init__(self, state, **kwargs):
        assert "spaces" not in kwargs
        assert "shape" not in kwargs
        kwargs["spaces"] = "c" * state.mesh.dim
        kwargs["shape"] = (3,)
        super().__init__(state, **kwargs)



# SPDX-License-Identifier: MIT

import hashlib
import importlib
import os
import pathlib
import pickle

from neuralmag.common import logging
from neuralmag.common.config import config


class CodeClass(object):
    def save_and_load_code(self, *args):
        # setup cache file name
        this_module = pathlib.Path(importlib.import_module(self.__module__).__file__)
        i = this_module.parent.parts[::-1].index("neuralmag")
        prefix = "_".join(
            (config.backend.name,) + this_module.parent.parts[-i:] + (this_module.stem,)
        )
        cache_file = f"{prefix}_{hashlib.md5(pickle.dumps(args)).hexdigest()}.py"
        cache_dir = os.getenv(
            "NM_CACHEDIR", pathlib.Path.home() / ".cache" / "neuralmag"
        )
        code_file_path = cache_dir / cache_file

        # generate code
        if not code_file_path.is_file():
            code_file_path.parent.mkdir(parents=True, exist_ok=True)
            # TODO check if _generate_code method exists
            logging.info_green(
                f"[{self.__class__.__name__}] Generate {config.backend.name} core methods"
            )
            code = str(self._generate_code(*args))
            with open(code_file_path, "w") as f:
                f.write(code)

        # import code
        module_spec = importlib.util.spec_from_file_location("code", code_file_path)
        self._code = importlib.util.module_from_spec(module_spec)
        module_spec.loader.exec_module(self._code)



# SPDX-License-Identifier: MIT

VERSION = "0.9.2"

import neuralmag.common.logging as logging
from neuralmag.common import *
from neuralmag.field_terms import *
from neuralmag.loggers import *
from neuralmag.solvers import *

logging.info_green(f"[NeuralMag] Version {VERSION}")

__all__ = common.__all__ + field_terms.__all__ + loggers.__all__ + solvers.__all__



# SPDX-License-Identifier: MIT

from neuralmag.common import config

__all__ = ["LLGSolver"]


def LLGSolver(state, scale_t=1e-9, parameters=None, **kwargs):
    """
    Factory method that returns a backend specific time-integrator object
    for the LLG (either :class:`LLGSolverTorch` or :class:`LLGSolverJAX`).

    :param state: The state used for the simulation
    :type state: :class:`State`
    :param scale_t: Internal scaling of time to improve numerical behavior
    :type scale_t: float, optional
    :param parameters: List a attribute names for the adjoint gradient computation.
                       Only required for optimization problems.
    :type parameters: list
    :param **kwargs: Additional backend specific solver options
    :type **kwargs: dict

    :Required state attributes:
        * **state.t** (*scalar*) The time in s
        * **state.h** (*nodal vector field*) The effective field in A/m
        * **state.m** (*nodal vector field*) The magnetization

    :Example:
        .. code-block::

            # create state with time and magnetization
            state = nm.State(nm.Mesh((10, 10, 10), (1e-9, 1e-9, 1e-9)))
            state.t = 0.0
            state.m = nm.VectorFunction(state).fill((1, 0, 0))

            # register constant Zeeman field as state.h
            nm.ExternalField(torch.Tensor((0, 0, 8e5))).register(state, "")

            # initiali LLGSolver
            llg = LLGSolver(state)

            # perform integration step
            llg.step(1e-12)

    """

    return config.backend.LLGSolver(
        state, scale_t=scale_t, parameters=parameters, **kwargs
    )



# SPDX-License-Identifier: MIT

from neuralmag.common.engine import N, Variable, dV
from neuralmag.field_terms.field_term import FieldTerm

__all__ = ["CubicAnisotropyField"]


class CubicAnisotropyField(FieldTerm):
    r"""
    Effective field contribution corresponding to the cubic anisotropy energy

    .. math::
      E = - \int_\Omega K \big( (\vec{m} \cdot
           \vec{e}_1)^2(\vec{m} \cdot
           \vec{e}_2)^2 + (\vec{m} \cdot
           \vec{e}_2)^2(\vec{m} \cdot
           \vec{e}_3)^2 + (\vec{m} \cdot
           \vec{e}_1)^2(\vec{m} \cdot
           \vec{e}_3)^2 \big) \dx

    with the anisotropy constant :math:`K` given in units of :math:`\text{J/m}^3`.

    :param n_gauss: Degree of Gauss quadrature used in the form compiler (defaults to 1).
    :type n_gauss: int

    :Required state attributes (if not renamed):
        * **state.material.Kc** (*cell scalar field*) The anisotropy constant in J/m^3
        * **state.material.Kc_axis1** (*cell vector field*) The first anisotropy axis as unit vector field
        * **state.material.Kc_axis2** (*cell vector field*) The second anisotropy axis as unit vector field
        * **state.material.Kc_axis3** (*cell vector field*) The third anisotropy axis as unit vector field
        * **state.material.Ms** (*cell scalar field*) The saturation magnetization in A/m

    **state.material.Kc_axis1**, **state.material.Kc_axis2**, and **state.material.Kc_axis3** should be
    orthogonal unit vectors.
    """
    default_name = "caniso"

    def __init__(self, n_gauss=1, **kwargs):
        super().__init__(n_gauss=n_gauss, **kwargs)

    @staticmethod
    def e_expr(m, dim):
        K = Variable("material__Kc", "c" * dim)
        axis1 = Variable("material__Kc_axis1", "c" * dim, (3,))
        axis2 = Variable("material__Kc_axis2", "c" * dim, (3,))
        axis3 = Variable("material__Kc_axis3", "c" * dim, (3,))
        return (
            -K
            * (
                m.dot(axis1) ** 2 * m.dot(axis2) ** 2
                + m.dot(axis2) ** 2 * m.dot(axis3) ** 2
                + m.dot(axis3) ** 2 * m.dot(axis1) ** 2
            )
            * dV(dim)
        )



# SPDX-License-Identifier: MIT

import importlib
import os

from neuralmag.common import logging

__all__ = ["config"]


class Config:
    def __init__(self):
        self._backend = None
        self._device = None
        self._dtype = None

        # public config keys
        self.torch = {
            "compile": os.getenv("NM_TORCH_COMPILE", "True").lower()
            in ("true", "1", "yes")
        }
        self.jax = {
            "jit": os.getenv("NM_JAX_JIT", "True").lower() in ("true", "1", "yes")
        }
        self.fem = {"n_gauss": 3}

    @property
    def backend(self):
        if self._backend is not None:
            return self._backend

        backend_name = os.getenv("NM_BACKEND", None)
        if backend_name is not None:
            self.backend = backend_name
        else:
            self.set_backend("jax", True)
            if self._backend is None:
                self.set_backend("torch", True)
            if self._backend is None:
                raise ImportError(
                    "Neither 'jax' nor 'torch' seems to be available. Use 'pip install \"neuralmag[torch]\"' or 'pip install \"neuralmag[jax]\"' to select one backend!"
                )

        return self._backend

    def set_backend(self, backend_name, graceful=False):
        if self._backend is not None:
            raise RuntimeError(
                "Backend is already set. Cannot switch backend at runtime."
            )
        if backend_name not in ["torch", "jax"]:
            raise ValueError(f"Unsupported backend: {backend_name}")

        self.backend_name = backend_name

        try:
            if backend_name == "torch":
                self._backend = importlib.import_module("neuralmag.backends.torch")
            if backend_name == "jax":
                self._backend = importlib.import_module("neuralmag.backends.jax")
            logging.info_green(f"[NeuralMag] Backend set to '{backend_name}'.")
        except ImportError:
            if not graceful:
                raise ImportError(
                    f"[NeuralMag] Could not load backend '{backend_name}'. Use 'pip install neuralmag[{backend_name}] to install."
                )

    @backend.setter
    def backend(self, backend_name):
        self.set_backend(backend_name)

    @property
    def device(self):
        if self._device is None:
            self.device = self.backend.default_device_str()
        return self._device

    @device.setter
    def device(self, device):
        self._device = self.backend.device_from_str(device)
        logging.info_green(f"[NeuralMag] Set default device to '{device}'.")

    @property
    def dtype(self):
        if self._dtype is None:
            self.dtype = self.backend.default_dtype_str()
        return self._dtype

    @dtype.setter
    def dtype(self, dtype):
        self._dtype = self.backend.dtype_from_str(dtype)
        logging.info_green(f"[NeuralMag] Set default dtype to '{dtype}'.")


config = Config()



# SPDX-License-Identifier: MIT

name = "jax"

import neuralmag.backends.jax.demag_field
from neuralmag.backends.jax.code_generation import *
from neuralmag.backends.jax.llg_solver_jax import *
from neuralmag.backends.jax.tensor_operations import *

LLGSolver = LLGSolverJAX



# SPDX-License-Identifier: MIT

import os
from collections.abc import Iterable

from ..common import logging
from .field_logger import FieldLogger
from .scalar_logger import ScalarLogger

__all__ = ["Logger"]


class Logger(object):
    """
    Combined Scalar- and Field-Logger class

    :param directory: The name of the log file
    :type directory: str
    :param scalars: List of state-attribute names of the scalars to be logged
    :type scalars: list
    :param fields: List of state-attribute names of the fields to be logged
    :type fields: list
    :param scalars_every: Write scalars every nth step
    :type scalars_every: int
    :param fields_every: Write fields every nth step
    :type fields_every: int

    :Example:
        .. code-block:: python

            # provide key strings with are available in state
            logger = Logger("data", ["m", "h_demag"], ["m"], fields_every=100)

            # Actually log fields
            state = State(mesh)
            logger.log(state)
    """

    def __init__(
        self, directory, scalars=[], fields=[], scalars_every=1, fields_every=1
    ):
        self.loggers = {}
        self._resume_time = None
        if len(scalars) > 0:
            self.loggers["scalars"] = ScalarLogger(
                os.path.join(directory, "log.dat"), scalars, every=scalars_every
            )
        if len(fields) > 0:
            self.loggers["fields"] = FieldLogger(
                os.path.join(directory, "fields.pvd"), fields, every=fields_every
            )

    def log(self, state):
        """
        Log simulation step

        :param state: The state to be logged
        :type state: :class:`State`
        """
        if state.t == self._resume_time:  # avoid logging directly after resume
            self._resume_time = None
            return
        for logger in self.loggers.values():
            logger.log(state)

    def resume(self, state):
        """
        Tries to resume from existing log files. If resume is possible,
        i.e. if at least one magnetization field has been logged before,
        the state object is updated with latest possible values of t
        and m and the different log files are aligned and resumed.
        If resume is not possible the state is not modified and the
        simulations starts from the beginning.

        :param state: The state to be resumed from
        :type state: :class:`State`
        """
        last_recorded_step = self.loggers["fields"].last_recorded_step()
        if last_recorded_step is None:
            logging.warning("Resume not possible. Start over.")
            return

        resumable_step = min(
            map(lambda logger: logger.resumable_step(), self.loggers.values())
        )
        assert resumable_step >= last_recorded_step + 1

        state.m, state.t = self.loggers["fields"].step_data(
            last_recorded_step, "m", state
        )
        logging.info_green(
            "Resuming from step %d (t = %g)." % (last_recorded_step, state.t)
        )

        for logger in self.loggers.values():
            logger.resume(last_recorded_step + 1)

        self._resume_time = state.t

    def is_resumable(self):
        """
        Returns True if logger can resume from log files.

        :return: True of resumable, False otherwise
        :rtype: bool
        """
        return self.loggers["fields"].last_recorded_step() is not None



# SPDX-License-Identifier: MIT

import jax
import jax.numpy as jnp
import numpy as np

from neuralmag.common import config, logging

float64 = jnp.float64
float32 = jnp.float32
integer = jnp.int32
Tensor = jax.Array

np = jnp
libs = {"jnp": jnp}


def device_from_str(device):
    return None


def device_for_state(device):
    logging.warning(
        f"[NeuralMag] JAX backend doesn't support setting device per State. Falling back to default device."
    )
    return None


def default_device_str():
    return jax.devices()[0]


def dtype_from_str(dtype):
    if dtype == "float64":
        jax.config.update("jax_enable_x64", True)

    return {"float64": float64, "float32": float32}[dtype]


def dtype_for_state(dtype):
    logging.warning(
        f"[NeuralMag] JAX backend doesn't support setting dtype per State. Falling back to default dtype."
    )
    return config.dtype


def default_dtype_str():
    return "float32"


def eps(dtype):
    return jnp.finfo(dtype).eps


def tensor(value, *, device=None, dtype=None):
    # if isinstance(value, jax.Array):
    #    if value.device != device:
    #        return jax.device_put(value, device)
    #    else:
    #        return value
    return jnp.array(value, dtype=dtype)


def zeros(shape, *, device=None, dtype=None, **kwargs):
    return jnp.zeros(shape, dtype=dtype, **kwargs)


def zeros_like(tensor):
    return jnp.zeros_like(tensor)


def arange(*args, device=None, dtype=None, **kwargs):
    return jnp.arange(*args, dtype=dtype, **kwargs)


def linspace(*args, device=None, dtype=None, **kwargs):
    return jnp.linspace(*args, dtype=dtype, **kwargs)


def meshgrid(*ranges, indexing="ij"):
    return jnp.meshgrid(*ranges, indexing="ij")


def to_numpy(array):
    return np.asarray(array)


def broadcast_to(array, shape):
    return jnp.broadcast_to(array, shape)


def tile(array, shape):
    return jnp.tile(array, shape)


def assign(target, source, idx):
    return target.at[idx].set(source)


def mean(tensor, axis=None):
    return jnp.mean(tensor, axis=axis)



# SPDX-License-Identifier: MIT

name = "torch"

import neuralmag.backends.torch.demag_field
from neuralmag.backends.torch.code_generation import *
from neuralmag.backends.torch.llg_solver_torch import *
from neuralmag.backends.torch.tensor_operations import *

LLGSolver = LLGSolverTorch



# SPDX-License-Identifier: MIT

from sympy.vector import curl

from neuralmag.common.engine import N, Variable, dV
from neuralmag.field_terms.field_term import FieldTerm

__all__ = ["BulkDMIField"]


class BulkDMIField(FieldTerm):
    r"""
    Effective field contribution for the micromagnetic bulk-DMI energy

    .. math::

      E = \int_\Omega D \vec{m} \cdot (\nabla \times \vec{m}) \dx

    with the DMI constant :math:`D` given in units of :math:`\text{J/m}^2`.

    :param n_gauss: Degree of Gauss quadrature used in the form compiler.
    :type n_gauss: int

    :Required state attributes (if not renamed):
        * **state.material.Db** (*cell scalar field*) The DMI constant in J/m^2
        * **state.material.Ms** (*cell scalar field*) The saturation magnetization in A/m
    """
    default_name = "bdmi"

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

    @staticmethod
    def e_expr(m, dim):
        D = Variable("material__Db", "c" * dim)
        return D * m.dot(curl(m)) * dV()



# SPDX-License-Identifier: MIT

from sympy.vector import divergence, gradient

from neuralmag.common.engine import N, Variable, dV
from neuralmag.field_terms.field_term import FieldTerm

__all__ = ["InterfaceDMIField"]


class InterfaceDMIField(FieldTerm):
    r"""
    Effective field contribution corresponding to the micromagnetic interface-DMI energy

    .. math::

      E = \int_\Omega D \Big[
         \vec{m} \cdot \nabla (\vec{e}_D \cdot \vec{m}) -
         (\nabla \cdot \vec{m}) (\vec{e}_D \cdot \vec{m})
         \Big] \dx

    with the DMI constant :math:`D` given in units of :math:`\text{J/m}^2`.

    :param n_gauss: Degree of Gauss quadrature used in the form compiler.
    :type n_gauss: int

    :Required state attributes (if not renamed):
        * **state.material.Di** (*cell scalar field*) The DMI constant in J/m^2
        * **state.material.Di_axis** (*cell vector field*) The DMI surface normal as unit vector field
        * **state.material.Ms** (*cell scalar field*) The saturation magnetization in A/m
    """
    default_name = "idmi"

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

    @staticmethod
    def e_expr(m, dim):
        D = Variable("material__Di", "c" * dim)
        axis = Variable("material__Di_axis", "c" * dim, (3,))
        return (
            D * (m.dot(gradient(m.dot(axis))) - divergence(m) * m.dot(axis)) * dV(dim)
        )



# SPDX-License-Identifier: MIT

import os
import xml.etree.cElementTree as ET
from xml.dom import minidom
from xml.etree import cElementTree

__all__ = ["FieldLogger"]


class FieldLogger(object):
    """
    Logger class for fields using PVD/VTI files from the visualization
    toolkit (VTK). The logger creates a single PVD file referencing the
    VTI snapshots for different simulation times saving spatially
    resolved field data.

    :param filename: The name of the log file
    :type filename: str
    :param fields: The fields to be written to the log file as a list of
                   attribute names of the state class
    :type fields: list
    :param every: Write field to log file every nth call
    :type every: int, optional

    :Example:
        .. code-block:: python

            # provide key strings which are available in state
            logger = FieldLogger("data/m.pvd", ["m", "h_demag"])

            # Actually log fields
            state = State(mesh)
            logger.log(state)
    """

    def __init__(self, filename, fields, every=1):
        # create directory if not existent
        if not os.path.dirname(filename) == "" and not os.path.exists(
            os.path.dirname(filename)
        ):
            try:
                os.makedirs(os.path.dirname(filename))
            except OSError as exc:  # Guard against race condition
                if exc.errno != errno.EEXIST:
                    raise

        filename, ext = os.path.splitext(filename)
        if ext != ".pvd":
            raise NameError("Only .pvd extention allowed")
        self._filename = filename
        self._every = every
        if isinstance(fields, str):
            fields = [fields]
        self._fields = fields
        self._i = 0
        self._i_start = 0
        self._xmlroot = cElementTree.Element(
            "VTKFile", type="Collection", version="0.1", byte_order="LittleEndian"
        )
        cElementTree.SubElement(self._xmlroot, "Collection")

    def log(self, state):
        """
        Log simulation step

        :param state: The state to be logged
        :type state: :class:`State`
        """
        self._i += 1
        if (self._i - 1) % self._every > 0:
            return
        if self._i <= self._i_start:
            return

        filename = "%s_%04d.vti" % (self._filename, self._i // self._every)
        state.write_vti(self._fields, filename)
        cElementTree.SubElement(
            self._xmlroot[0],
            "DataSet",
            timestep=str(state.t.tolist()),
            file=os.path.basename(filename),
        )
        with open(self._filename + ".pvd", "w") as fd:
            fd.write(
                minidom.parseString(
                    " ".join(
                        cElementTree.tostring(self._xmlroot)
                        .decode()
                        .replace("\n", "")
                        .split()
                    ).replace("> <", "><")
                ).toprettyxml(indent="  ")
            )

    def reset(self):
        """
        Reset the internal step counter
        """
        self._i = 0

    def resumable_step(self):
        """
        Returns the first step that can be written when resuming, e.g. if the
        logger logs every 10th step and the first (i = 0) step was already
        logged, the result is 10.

        :return: The step number
        :rtype: int
        """
        try:
            xml = cElementTree.parse(self._filename + ".pvd").getroot()
            return len(list(xml.find("Collection"))) * self._every
        except IOError:
            return 0

    def last_recorded_step(self):
        """
        Returns the number of the last step logged and ``None`` if no
        step was yet logged.

        :return: Number of the last step recorded
        :rtype: int/None
        """
        result = (self.resumable_step() // self._every - 1) * self._every
        if result < 0:
            return None
        else:
            return result

    def step_data(self, i, field, state):
        """
        Returns field and time to a given step number.

        :param i: The step number
        :type i: int
        :param field: The name of the field to be read
        :type field: str
        :param state: The state used for the read
        :type state: :class:`State`
        :return: The field as a function
        :rtype: :class:`Function`
        """
        if i % self._every > 0:
            raise Exception()

        xml = cElementTree.parse(self._filename + ".pvd").getroot()
        item = list(xml.find("Collection"))[i // self._every]
        field = state.read_vti(
            os.path.join(os.path.dirname(self._filename), item.attrib["file"]), field
        )
        return field, float(item.attrib["timestep"])

    def resume(self, i):
        """
        Try to resume existing log file from log step i. The log file
        is truncated accordingly.

        :param i: The log step to resume from
        :type i: int
        """
        self._i = i
        self._i_start = self.last_recorded_step() + 1
        self._xmlroot = cElementTree.parse(self._filename + ".pvd").getroot()



# SPDX-License-Identifier: MIT

import logging

__all__ = [
    "set_log_level",
    "debug",
    "warning",
    "error",
    "info",
    "info_green",
    "info_blue",
]

# create magnum.fe logger
logger = logging.getLogger("NeuralMag")

handler = logging.StreamHandler()
handler.setFormatter(
    logging.Formatter(
        fmt="%(asctime)s %(name)s:%(levelname)s %(message)s",
        datefmt="%Y-%m-%d %H:%M:%S",
    )
)
logger.addHandler(handler)

logger.setLevel(logging.INFO)

info = logger.info

RED = "\033[1;37;31m%s\033[0m"
BLUE = "\033[1;37;34m%s\033[0m"
GREEN = "\033[1;37;32m%s\033[0m"
CYAN = "\033[1;37;36m%s\033[0m"


def debug(message, *args, **kwargs):
    logger.debug(CYAN % message, *args, **kwargs)


def warning(message, *args, **kwargs):
    logger.warning(RED % message, *args, **kwargs)


def error(message, *args, **kwargs):
    logger.error(RED % message, *args, **kwargs)


def info_green(message, *args, **kwargs):
    info(GREEN % message, *args, **kwargs)


def info_blue(message, *args, **kwargs):
    info(BLUE % message, *args, **kwargs)


def set_log_level(level):
    """
    Set the log level of magnum.np specific logging messages.
    Defaults to :code:`INFO = 20`.

    *Arguments*
      level (:class:`int`)
        The log level
    """
    logger.setLevel(level)



# SPDX-License-Identifier: MIT

import json
import re
from functools import reduce
from itertools import product

import sympy as sp
import sympy.vector as sv
from scipy.special import p_roots
from tqdm import tqdm

N = sv.CoordSys3D("N")
cs_dx = sp.symbols("_dx[0]_ _dx[1]_ _dx[2]_", real=True, positive=True)
cs_x = [N.x, N.y, N.z]
cs_e = [N.i, N.j, N.k]


def dX(**kwargs):
    r"""
    Generic integration measure

    :param \**kwargs: Arbitrary meta information
    :return: The integration measure
    :rtype: sympy.Expr
    """
    # TODO implement singleton
    return sp.Symbol(f"_dX:{json.dumps(kwargs)}_")


def dV(dim=3, region="rho", **kwargs):
    r"""
    Volume integral measure

    :param dim: Dimension of the mesh to integrate over.
    :type dim: int
    :param region: name of the cell function that acts as a region indicator
    :type region: str
    :param \**kwargs: Additional meta information
    :return: The integration measure
    :rtype: sympy.Expr
    """
    rho = Variable(region, "c" * dim)
    return rho * dX(dims=[None, None, None], **kwargs)


def dA(dim=3, normal=2, region="rhoxy", idx=":", **kwargs):
    r"""
    Volume integral measure

    :param dim: Dimension of the mesh to integrate over.
    :type dim: int
    :param region: name of the cell function that acts as a region indicator
    :type region: str
    :param \**kwargs: Additional meta information
    :return: The integration measure
    :type: sympy.Expr
    """
    assert dim == 3
    spaces = ["c"] * 3
    spaces[normal] = "n"
    rho = Variable(region, "".join(spaces))
    dims = [None, None, None]
    dims[normal] = idx
    return rho * dX(dims=dims, **kwargs)


def Variable(name, spaces, shape=()):
    r"""
    Symbolic representation of a field given as SymPy expression.

    :param name: The name of the field
    :type name: str
    :param spaces: The function spaces of the field in the principal coordinate
                   directions given as string with 'c' representing a cell-based
                   discretization and 'n' representing a node-based discretization.
    :type spaces: str
    :param shape: The shape (dimension) of the field, e.g. () for a scalar field and
                  (3,) for a vector field
    :type shape: tuple
    :return: The variable
    :rtype: sympy.Expr
    """
    result = []
    for idx in product(*[{"n": [0, 1], "c": [None]}[s] for s in spaces]):
        phi = 1.0
        for i, j in enumerate(idx):
            if j is not None:
                phi *= 1 - cs_x[i] / cs_dx[i] + 2 * j * cs_x[i] / cs_dx[i] - j
        if shape == ():
            result.append(
                sp.Symbol(f"_{name}:{spaces}:{shape}:{list(idx)}_", real=True) * phi
            )
        elif shape == (3,):
            for l in range(3):
                result.append(
                    sp.Symbol(f"_{name}:{spaces}:{shape}:{list(idx) + [l]}_", real=True)
                    * phi
                    * cs_e[l]
                )
        else:
            raise Exception("Shape not supported")
    return reduce(lambda x, y: x + y, result)


def integrate(expr, dims, n=3):
    x, w = p_roots(n)

    integrand = expr
    for i, dim in enumerate(dims):
        if dim is None:
            integral = 0
            for j in range(n):
                integral += (
                    w[j]
                    * cs_dx[i]
                    / 2
                    * integrand.subs(cs_x[i], (1 + x[j]) * cs_dx[i] / 2)
                )
        else:
            integral = integrand.subs(cs_x[i], 0.0)
        integrand = integral

    return integral


def compile_functional(expr, n_gauss=3):
    # extract all integral measures with parameters and check consistency
    measure_symbols = [s for s in expr.free_symbols if re.match(r"^_dX:(.*)_$", s.name)]
    integrals = sp.collect(expr, measure_symbols, exact=True, evaluate=False)
    assert 1 not in integrals

    cmds = []
    variables = {"dx"}
    for symb in measure_symbols:
        match = re.match(r"^_dX:(.*)_$", symb.name)
        args = json.loads(match[1])

        # integrate
        # TODO use | operator for python 3.9
        iexpr = integrate(integrals[symb], **{**{"n": n_gauss}, **args})

        # skip zero integrals
        if iexpr.is_zero:
            continue

        # find all named symbols (fields)
        symbs = [
            symb
            for symb in iexpr.free_symbols
            if re.match(r"^_(.*:.*:.*:.*)_$", symb.name)
        ]

        if len(symbs) == 0:
            raise Exception("Need at least one variable to integrate.")

        ## try to reduce multiplications of fields for better performance
        ## the switch on n_gaus is purely heuristic and was introduced for the cubic anisotropy
        if n_gauss == 1:
            cmd = str(sp.collect(sp.factor_terms(iexpr), symbs))
        else:
            cmd = str(sp.collect(sp.factor_terms(sp.expand(iexpr)), symbs))

        # retrieve topological dimension from first symbol
        match = re.match(r"^_(.*:.*:.*:.*)_$", symbs[0].name)
        shape, idx = [eval(x) for x in match[1].split(":")[2:]]
        dim = len(idx) - len(shape)

        for symb in symbs:
            match = re.match(r"^_(.*:.*:.*:.*)_$", symb.name)
            name, spaces = match[1].split(":")[:2]
            shape, idx = [eval(x) for x in match[1].split(":")[2:]]

            variables.add(name)

            sidx = []
            for i, space in enumerate(spaces):
                if space == "n":
                    if args["dims"][i] is None:
                        sidx.append([":-1", "1:"][idx[i]])
                    else:
                        sidx.append(str(args["dims"][i]))
                        if isinstance(args["dims"][i], str):
                            variables.add(args["dims"][i])
                elif space == "c":
                    if args["dims"][i] is None:
                        sidx.append(":")
                    else:
                        raise Exception("Use node discretization in normal direction.")

            if shape == (3,):
                sidx.append(str(idx[-1]))

            # contract leading sequence of ":,: to ...
            arr_idx = re.sub(r"^(:,)*:($|,)", r"...\2", ",".join(sidx))
            cmd = cmd.replace(symb.name, f"{name}[{arr_idx}]")

        args["cmd"] = re.sub(r"_(dx\[\d\])_", r"\1", cmd)
        cmds.append(args)

    return cmds, variables


def linear_form_cmds(expr, n_gauss=3):
    cmds = []
    v = {}

    # collect all test functions in expr
    for symb in sorted(list(expr.free_symbols), key=lambda s: s.name):
        match = re.match(r"^_v:(.*:.*:.*)_$", symb.name)
        if match:
            v[symb] = match[1].split(":")
            v[symb][1:] = [eval(x) for x in v[symb][1:]]

    # retrieve topological dimension from first symbol
    _, shape, idx = next(iter(v.values()))
    dim = len(idx) - len(shape)

    # process test functions
    variables = set()
    for vsymb in tqdm(v, desc="Generating..."):
        vexpr = expr.xreplace(dict([(s, 1.0) if s == vsymb else (s, 0.0) for s in v]))
        terms, vvars = compile_functional(vexpr, n_gauss)
        variables = variables.union(vvars)
        vspaces, vshape, vidx = v[vsymb]

        for term in terms:
            # TODO why call it term here and args in compile_function?
            sidx = []
            for i, space in enumerate(vspaces):
                if space == "n":
                    if term["dims"][i] is None:
                        sidx.append([":-1", "1:"][vidx[i]])
                    else:
                        sidx.append(str(term["dims"][i]))
                elif space == "c":
                    if term["dims"][i] is None:
                        sidx.append(":")
                    else:
                        raise Exception("Use node discretization in normal direction.")

            if shape == (3,):
                sidx.append(str(vidx[-1]))

            cmds.append((",".join(sidx), term["cmd"]))

    return cmds, variables


def gateaux_derivative(expr, var):
    r"""
    Compute the Gateaux derivative (variation) of a functional with respect to
    a given variable.

    :param expr: Functional to be derived
    :type expr: sympy.Expr
    :param var: The variable used for the derivative
    :type var: :class:`Variable`
    :return: The resulting linear form
    :rtype: sympy.Expr
    """
    result = []
    for symb in var.free_symbols:
        if not hasattr(symb, "name") or not re.match(r"^_(.*:.*:.*:.*)_$", symb.name):
            continue
        v = sp.Symbol(re.sub(r"^_.*:(.*:.*:.*_)$", r"_v:\1", symb.name))
        result.append(v * expr.diff(symb))
    return reduce(lambda x, y: x + y, result)



# SPDX-License-Identifier: MIT

import os
from time import time

from scipy import constants

from neuralmag.common import config
from neuralmag.common.engine import Variable, dV
from neuralmag.field_terms.field_term import FieldTerm

__all__ = ["DemagField"]


class DemagField(FieldTerm):
    r"""
    Effective field contribution corresponding to the demagnetization field (also referred to as magnetostatic field or stray field).
    The demagnetization field is computed from the scalar potential :math:`u` as

    .. math::

        \vec{H}_\text{demag} = - \nabla u

    with :math:`u` being calculated by the Poisson equation

    .. math::

      \Delta u = \nabla \cdot (M_s \vec{m})

    with open boundary conditions.


    :param p: Distance threshhold at which the demag tensor is approximated
              by a dipole field given in numbers of cells. Defaults to 20.
    :type p: int
    :param n_gauss: Degree of Gauss quadrature used in the form compiler.
    :type n_gauss: int

    :Required state attributes (if not renamed):
        * **state.material.Ms** (*cell scalar field*) The saturation magnetization in A/m
    """
    default_name = "demag"
    h = None

    def __init__(self, p=20, **kwargs):
        super().__init__(**kwargs)
        self._p = p

    def register(self, state, name=None):
        super().register(state, name)
        if state.mesh.dim == 2:
            setattr(
                state,
                self.attr_name("h", name),
                (config.backend.demag_field.h2d, "nn", (3,)),
            )
        elif state.mesh.dim == 3:
            setattr(
                state,
                self.attr_name("h", name),
                (config.backend.demag_field.h3d, "nnn", (3,)),
            )
        else:
            raise
        # fix reference to h_demag in E_demag if suffix is changed
        if name is not None:
            wrapped = state.wrap_func(self.E, {"h_demag": self.attr_name("h", name)})
            setattr(state, self.attr_name("E", name), wrapped)
        config.backend.demag_field.init_N(state, self._p)

    @staticmethod
    def e_expr(m, dim):
        rho = Variable("rho", "c" * dim)
        Ms = Variable("material__Ms", "c" * dim)
        h_demag = Variable("h_demag", "n" * dim, (3,))
        return -0.5 * constants.mu_0 * Ms * m.dot(h_demag) * dV()



# SPDX-License-Identifier: MIT

from functools import reduce

from neuralmag.common import logging

__all__ = ["Mesh"]


class Mesh(object):
    r"""
    Regular cuboid mesh for nodal finite-differences. If number of cells in
    all principal directions are provided, a 3D mesh is created with
    :math:`(n_1 + 1) \times (n_2 + 1) \times (n_3 + 1)` cells.
    If only :math:`n_1` and :math:`n_2` is provided, a 2D mesh is created
    with a single layer of nodes, but finite thickness.
    If only :math:`n_1` is provided, a 1D mesh is created with a single line
    of nodes, but finite thickness and width.

    :param n: Number of cells in principal directions
    :type n: tuple
    :param dx: Cell-size in principal directions in m
    :type dx: tuple
    :param origin: Coordinate of the bottom-left corner of the mesh
    :type origin: tuple

    :Example:
        .. code-block::

            # 3D with 1 cell thickness, leading to 2 nodes in z-direction
            mesh_3d = Mesh((100, 25, 1), (5e-9, 5e-9, 3e-9))

            # 2D with 1 cell thickness, leading to 1 nodes in z-direction
            mesh_2d = Mesh((100, 25), (5e-9, 5e-9, 3e-9))
    """

    def __init__(self, n, dx, origin=(0, 0, 0)):
        self.n = tuple(n)
        self.dim = len(n)
        self.dx = tuple(dx)
        self.origin = tuple(origin)
        logging.info_green(
            f"[Mesh] {self.dim}D, {' x '.join([str(x) for x in self.n])} (size = {' x '.join(['{:,g}'.format(x) for x in self.dx])})"
        )

    @property
    def cell_volume(self):
        """
        The cell volume
        """
        return self.dx[0] * self.dx[1] * self.dx[2]

    @property
    def volume(self):
        """
        The total volume of the mesh
        """
        return self.num_cells * self.cell_volume

    @property
    def num_cells(self):
        """
        The total number of simulation cells
        """
        return reduce(lambda x, y: x * y, self.n)

    @property
    def num_nodes(self):
        """
        The total number of simulation nodes
        """
        # TODO better use a loop here?
        if self.dim == 3:
            return (self.n[0] + 1) * (self.n[1] + 1) * (self.n[2] + 1)
        elif self.dim == 2:
            return (self.n[0] + 1) * (self.n[1] + 1)
        elif self.dim == 1:
            return self.n[0] + 1
        else:
            raise RuntimeError(f"Mesh dimension '{self.dim}' not supported")

    def __str__(self):
        return f"{'x'.join(str(x) for x in self.n)}_{self.dx[0]:g}x{self.dx[1]:g}x{self.dx[2]:g}"



# SPDX-License-Identifier: MIT

import types

from neuralmag.common import logging
from neuralmag.field_terms.field_term import FieldTerm

__all__ = ["TotalField"]


class TotalField(FieldTerm):
    r"""
    This class combines multiple field terms into a single total field term by
    adding up their effective fields and energies.

    :param \*field_names: The names of the effective field contributions

    :Example:
        .. code-block::

            state = nm.State(nm.Mesh((10, 10, 10), (1e-9, 1e-9, 1e-9)))

            nm.ExchangeField().register(state, "exchange")
            nm.DemagField().register(state, "demag")
            nm.ExternalField(h_ext).register(state, "external")
            nm.TotalField("exchange", "demag", "external").register(state)

            # Compute total field and energy
            h = state.h
            E = state.E
    """
    default_name = ""

    def __init__(self, *field_names):
        self._field_names = field_names

    def register(self, state, name=None):
        code = f"def h_total({', '.join([self.attr_name('h', name) for name in self._field_names])}):\n"
        code += (
            "    return"
            f" {' + '.join([self.attr_name('h', name) for name in self._field_names])}"
        )
        compiled_code = compile(code, "<string>", "exec")
        h_func = types.FunctionType(compiled_code.co_consts[0], {}, "h_total")

        code = f"def E_total({', '.join([self.attr_name('E', name) for name in self._field_names])}):\n"
        code += (
            "    return"
            f" {' + '.join([self.attr_name('E', name) for name in self._field_names])}"
        )
        compiled_code = compile(code, "<string>", "exec")
        E_func = types.FunctionType(compiled_code.co_consts[0], {}, "E_total")

        logging.info_green(
            f"[{self.__class__.__name__}] Register state methods (field:"
            f" '{self.attr_name('h', name)}', energy: '{self.attr_name('E', name)}')"
        )
        setattr(state, self.attr_name("h", name), (h_func, "n" * state.mesh.dim, (3,)))
        setattr(state, self.attr_name("E", name), E_func)



# SPDX-License-Identifier: MIT

import numpy as np
import torch
import torch.fft
from scipy import constants
from torch import abs, asinh, atan, log, sqrt

from neuralmag.common import logging

complex_dtype = {
    torch.float: torch.complex,
    torch.float32: torch.complex64,
    torch.float64: torch.complex128,
}


def f(x, y, z):
    x, y, z = abs(x), abs(y), abs(z)
    x2, y2, z2 = x**2, y**2, z**2
    r = sqrt(x2 + y2 + z2)
    result = 1.0 / 6.0 * (2 * x2 - y2 - z2) * r
    result += (y / 2.0 * (z2 - x2) * asinh(y / sqrt(x2 + z2))).nan_to_num(
        posinf=0, neginf=0
    )
    result += (z / 2.0 * (y2 - x2) * asinh(z / sqrt(x2 + y2))).nan_to_num(
        posinf=0, neginf=0
    )
    result -= (x * y * z * atan(y * z / (x * r))).nan_to_num(posinf=0, neginf=0)
    return result


def g(x, y, z):
    z = abs(z)
    x2, y2, z2 = x**2, y**2, z**2
    r = sqrt(x2 + y2 + z2)
    result = -x * y * r / 3.0
    result += (x * y * z * asinh(z / sqrt(x2 + y2))).nan_to_num(posinf=0, neginf=0)
    result += (y / 6.0 * (3.0 * z2 - y2) * asinh(x / sqrt(y2 + z2))).nan_to_num(
        posinf=0, neginf=0
    )
    result += (x / 6.0 * (3.0 * z2 - x2) * asinh(y / sqrt(x2 + z2))).nan_to_num(
        posinf=0, neginf=0
    )
    result -= (z**3 / 6.0 * atan(x * y / (z * r))).nan_to_num(posinf=0, neginf=0)
    result -= (z * y2 / 2.0 * atan(x * z / (y * r))).nan_to_num(posinf=0, neginf=0)
    result -= (z * x2 / 2.0 * atan(y * z / (x * r))).nan_to_num(posinf=0, neginf=0)
    return result


def F1(func, x, y, z, dz, dZ):
    return (
        func(x, y, z + dZ)
        - func(x, y, z)
        - func(x, y, z - dz + dZ)
        + func(x, y, z - dz)
    )


def F0(func, x, y, z, dy, dY, dz, dZ):
    return (
        F1(func, x, y + dY, z, dz, dZ)
        - F1(func, x, y, z, dz, dZ)
        - F1(func, x, y - dy + dY, z, dz, dZ)
        + F1(func, x, y - dy, z, dz, dZ)
    )


def newell(func, x, y, z, dx, dy, dz, dX, dY, dZ):
    ret = (
        F0(func, x, y, z, dy, dY, dz, dZ)
        - F0(func, x - dx, y, z, dy, dY, dz, dZ)
        - F0(func, x + dX, y, z, dy, dY, dz, dZ)
        + F0(func, x - dx + dX, y, z, dy, dY, dz, dZ)
    )
    return -ret / (4.0 * np.pi * dx * dy * dz)


def dipole_f(x, y, z, dx, dy, dz, dX, dY, dZ):
    z = z + dZ / 2.0 - dz / 2.0  # diff of cell centers for non-equidistant demag
    result = (2.0 * x**2 - y**2 - z**2) * pow(
        x**2 + y**2 + z**2, -5.0 / 2.0
    )
    result[0, 0, 0] = 0.0
    return result * dx * dy * dz / (4.0 * np.pi)


def dipole_g(x, y, z, dx, dy, dz, dX, dY, dZ):
    z = z + dZ / 2.0 - dz / 2.0  # diff of cell centers for non-equidistant demag
    result = 3.0 * x * y * pow(x**2 + y**2 + z**2, -5.0 / 2.0)
    result[0, 0, 0] = 0.0
    return result * dx * dy * dz / (4.0 * np.pi)


def demag_f(x, y, z, dx, dy, dz, dX, dY, dZ, p):
    res = dipole_f(x, y, z, dx, dy, dz, dX, dY, dZ)
    near = (x**2 + y**2 + z**2) / max(
        dx**2 + dy**2 + dz**2, dX**2 + dY**2 + dZ**2
    ) < p**2
    res[near] = newell(f, x[near], y[near], z[near], dx, dy, dz, dX, dY, dZ)
    return res


def demag_g(x, y, z, dx, dy, dz, dX, dY, dZ, p):
    res = dipole_g(x, y, z, dx, dy, dz, dX, dY, dZ)
    near = (x**2 + y**2 + z**2) / max(
        dx**2 + dy**2 + dz**2, dX**2 + dY**2 + dZ**2
    ) < p**2
    res[near] = newell(g, x[near], y[near], z[near], dx, dy, dz, dX, dY, dZ)
    return res


def h_cell(N_demag, mcell, material__Ms, rho):
    dim = [i for i in range(3) if mcell.shape[i] > 1]
    s = [mcell.shape[i] * 2 for i in dim]

    hx = torch.zeros_like(N_demag[0][0], dtype=complex_dtype[mcell.dtype])
    hy = torch.zeros_like(N_demag[0][0], dtype=complex_dtype[mcell.dtype])
    hz = torch.zeros_like(N_demag[0][0], dtype=complex_dtype[mcell.dtype])
    for ax in range(3):
        m_pad_fft1D = torch.fft.rfftn(
            rho * material__Ms * mcell[:, :, :, ax], dim=dim, s=s
        )
        hx += N_demag[0][ax] * m_pad_fft1D
        hy += N_demag[1][ax] * m_pad_fft1D
        hz += N_demag[2][ax] * m_pad_fft1D

    hx = torch.fft.irfftn(hx, dim=dim)
    hy = torch.fft.irfftn(hy, dim=dim)
    hz = torch.fft.irfftn(hz, dim=dim)

    return torch.stack(
        [
            hx[: mcell.shape[0], : mcell.shape[1], : mcell.shape[2]],
            hy[: mcell.shape[0], : mcell.shape[1], : mcell.shape[2]],
            hz[: mcell.shape[0], : mcell.shape[1], : mcell.shape[2]],
        ],
        dim=3,
    )


def h2d(N_demag, m, material__Ms, rho):
    mcell = (m[1:, 1:, :] + m[:-1, 1:, :] + m[1:, :-1, :] + m[:-1, :-1, :]).unsqueeze(
        -2
    ) / 4.0

    # TODO behavior for unsqueezed scalars seems OK, but not so elegant?
    Ms_hcell = (rho * material__Ms).unsqueeze(-1) * h_cell(
        N_demag, mcell, material__Ms.unsqueeze(-1), rho.unsqueeze(-1)
    ).squeeze(-2)

    h = torch.zeros(m.shape, dtype=m.dtype, device=m.device)
    h[:-1, :-1] += Ms_hcell
    h[:-1, 1:] += Ms_hcell
    h[1:, :-1] += Ms_hcell
    h[1:, 1:] += Ms_hcell

    mass = torch.zeros(h.shape[:-1], dtype=h.dtype, device=h.device)
    mass[:-1, :-1] += rho * material__Ms
    mass[:-1, 1:] += rho * material__Ms
    mass[1:, :-1] += rho * material__Ms
    mass[1:, 1:] += rho * material__Ms

    return h / mass.unsqueeze(-1)


def h3d(N_demag, m, material__Ms, rho):
    mcell = (
        +m[1:, 1:, 1:, :]
        + m[:-1, 1:, 1:, :]
        + m[1:, :-1, 1:, :]
        + m[:-1, :-1, 1:, :]
        + m[1:, 1:, :-1, :]
        + m[:-1, 1:, :-1, :]
        + m[1:, :-1, :-1, :]
        + m[:-1, :-1, :-1, :]
    ) / 8.0

    Ms_hcell = (rho * material__Ms).unsqueeze(-1) * h_cell(
        N_demag, mcell, material__Ms, rho
    )

    h = torch.zeros(m.shape, dtype=m.dtype, device=m.device)
    h[:-1, :-1, :-1] += Ms_hcell
    h[:-1, :-1, 1:] += Ms_hcell
    h[:-1, 1:, :-1] += Ms_hcell
    h[:-1, 1:, 1:] += Ms_hcell
    h[1:, :-1, :-1] += Ms_hcell
    h[1:, :-1, 1:] += Ms_hcell
    h[1:, 1:, :-1] += Ms_hcell
    h[1:, 1:, 1:] += Ms_hcell

    mass = torch.zeros(h.shape[:-1], dtype=h.dtype, device=h.device)
    mass[:-1, :-1, :-1] += rho * material__Ms
    mass[:-1, :-1, 1:] += rho * material__Ms
    mass[:-1, 1:, :-1] += rho * material__Ms
    mass[:-1, 1:, 1:] += rho * material__Ms
    mass[1:, :-1, :-1] += rho * material__Ms
    mass[1:, :-1, 1:] += rho * material__Ms
    mass[1:, 1:, :-1] += rho * material__Ms
    mass[1:, 1:, 1:] += rho * material__Ms

    return h / mass.unsqueeze(-1)


def init_N_component(state, perm, func, p):
    n = state.mesh.n + tuple([1] * (3 - state.mesh.dim))
    dx = np.array(state.mesh.dx)
    dx /= dx.min()  # rescale dx to avoid NaNs when using single precision

    shape = [i * 2 if i > 1 else i for i in n]
    ij = [
        torch.fft.fftfreq(n, 1 / n).to(dtype=torch.float64, device=state.device)
        for n in shape
    ]  # local indices
    ij = torch.meshgrid(*ij, indexing="ij")
    x, y, z = [ij[ind] * dx[ind] for ind in perm]
    Lx = [n[ind] * dx[ind] for ind in perm]
    dx = [dx[ind] for ind in perm]

    # TODO enable pseudo PBCs
    # offsets = [state.arange(-state.mesh.pbc[ind], state.mesh.pbc[ind]+1) for ind in perm] # offset of pseudo PBC images
    # offsets = torch.stack(torch.meshgrid(*offsets, indexing="ij"), dim=-1).flatten(end_dim=-2)
    # Nc = state.zeros(shape)
    # for offset in offsets:
    #    Nc += func(x + offset[0]*Lx[0], y + offset[1]*Lx[1], z + offset[2]*Lx[2], *dx, *dx, self._p)
    Nc = func(x, y, z, *dx, *dx, p)

    dim = [i for i in range(3) if n[i] > 1]
    if len(dim) > 0:
        Nc = torch.fft.rfftn(Nc, dim=dim)
    return Nc.real.clone()


def init_N(state, p):
    logging.info_green(f"[DemagField]: Set up demag tensor")

    Nxx = init_N_component(state, [0, 1, 2], demag_f, p).to(dtype=state.dtype)
    Nxy = init_N_component(state, [0, 1, 2], demag_g, p).to(dtype=state.dtype)
    Nxz = init_N_component(state, [0, 2, 1], demag_g, p).to(dtype=state.dtype)
    Nyy = init_N_component(state, [1, 2, 0], demag_f, p).to(dtype=state.dtype)
    Nyz = init_N_component(state, [1, 2, 0], demag_g, p).to(dtype=state.dtype)
    Nzz = init_N_component(state, [2, 0, 1], demag_f, p).to(dtype=state.dtype)

    state.N_demag = [[Nxx, Nxy, Nxz], [Nxy, Nyy, Nyz], [Nxz, Nyz, Nzz]]



# SPDX-License-Identifier: MIT

import jax

from neuralmag.common import config


def linear_form_code(form, n_gauss=3):
    r"""
    Generate PyTorch function for the evaluation of a given linear form.

    :param form: The linear form
    :type form: sympy.Expr
    :param n_gauss: Degree of Gauss integration
    :type n_gauss: int
    :return: The Python code of the PyTorch function
    :rtype: str
    """
    cmds, variables = linear_form_cmds(form, n_gauss)
    code = CodeBlock()
    with code.add_function("L", ["result"] + sorted(list(variables))) as f:
        for cmd in cmds:
            f.add_to("result", cmd[0], cmd[1])

    return code


def functional_code(form, n_gauss=3):
    r"""
    Generate PyTorch function for the evaluation of a given functional form.

    :param form: The functional
    :type form: sympy.Expr
    :param n_gauss: Degree of Gauss integration
    :type n_gauss: int
    :return: The Python code of the PyTorch function
    :rtype: str
    """
    terms, variables = compile_functional(form, n_gauss)
    code = CodeBlock()
    with code.add_function("M", sorted(list(variables))) as f:
        f.retrn_sum(*[term["cmd"] for term in terms])

    return code


def compile(func):
    if config.jax["jit"]:
        return jax.jit(func)
    else:
        return func


class CodeFunction(object):
    def __init__(self, block, name, variables):
        self._block = block
        self._name = name
        self._variables = variables

    def __enter__(self):
        self._code = f"def {self._name}({', '.join(self._variables)}):\n"
        return self

    def __exit__(self, type, value, traceback):
        if type is not None:
            return False
        self._block.add(self._code)
        self._block.add("\n")
        return True

    @staticmethod
    def sum(*terms):
        return " + ".join([f"({term}).sum()" for term in terms])

    def add_line(self, code):
        self._code += f"    {code}\n"

    def assign(self, lhs, rhs, index=None):
        if index is None:
            self.add_line(f"{lhs} = {rhs}")
        else:
            self.add_line(f"{lhs} = {lhs}.at[{index}].set({rhs})")

    def assign_sum(self, lhs, *terms, index=None):
        self.assign(lhs, self.sum(*terms), index)

    def zeros_like(self, var, src, shape=None):
        if shape is None:
            self.add_line(f"{var} = jnp.zeros_like({src})")
        else:
            self.add_line(f"{var} = jnp.zeros({shape}, dtype = {src}.dtype)")

    def add_to(self, var, idx, rhs):
        self.add_line(f"{var} = {var}.at[{idx}].add({rhs})")

    def retrn(self, code):
        self.add_line(f"return {code}")

    def retrn_sum(self, *terms):
        self.add_line(f"return {self.sum(*terms)}")

    def retrn_expanded(self, code, shape):
        self.add_line(f"return jnp.broadcast_to({code}, {shape})")


class CodeBlock(object):
    def __init__(self, plain=False):
        if plain:
            self._code = ""
        else:
            self._code = "import jax.numpy as jnp\n\n"

    def add_function(self, name, variables):
        return CodeFunction(self, name, variables)

    def add(self, code):
        self._code += code

    def __str__(self):
        return self._code



# SPDX-License-Identifier: MIT

from neuralmag.common import engine
from neuralmag.common.code_class import *
from neuralmag.common.config import config
from neuralmag.common.function import *
from neuralmag.common.logging import *
from neuralmag.common.mesh import *
from neuralmag.common.state import *

__all__ = (
    ["config", "engine"]
    + logging.__all__
    + function.__all__
    + mesh.__all__
    + state.__all__
)



# SPDX-License-Identifier: MIT

import os

import torch

from neuralmag.common import logging

float64 = torch.float64
float32 = torch.float32
integer = torch.int
Tensor = torch.Tensor

np = torch
libs = {"torch": torch}


def device_from_str(device):
    return torch.device(device)


def device_for_state(device):
    return device_from_str(device)


def default_device_str():
    return (
        f"cuda:{os.environ.get('CUDA_DEVICE', '0')}"
        if torch.cuda.is_available()
        else "cpu"
    )


def dtype_from_str(dtype):
    return {"float64": float64, "float32": float32}[dtype]


def dtype_for_state(dtype):
    return dtype_from_str(dtype)


def default_dtype_str():
    return "float32"


def eps(dtype):
    return torch.finfo(dtype).eps


def tensor(value, *, device=None, dtype=None, requires_grad=False):
    if isinstance(value, torch.Tensor):
        if value.device != device:
            return value.to(device)
        else:
            return value
    return torch.tensor(value, device=device, dtype=dtype, requires_grad=requires_grad)


def zeros(shape, *, device=None, dtype=None, **kwargs):
    return torch.zeros(shape, device=device, dtype=dtype, **kwargs)


def zeros_like(tensor):
    return torch.zeros_like(tensor)


def arange(*args, device=None, dtype=None, **kwargs):
    return torch.arange(*args, device=device, dtype=dtype, **kwargs)


def linspace(*args, device=None, dtype=None, **kwargs):
    return torch.linspace(*args, device=device, dtype=dtype, **kwargs)


def meshgrid(*ranges, indexing="ij"):
    return torch.meshgrid(*ranges, indexing="ij")


def to_numpy(array):
    return array.detach().cpu().numpy()


def broadcast_to(array, shape):
    return array.expand(shape)


def tile(array, shape):
    return torch.tile(array, shape)


def assign(target, source, idx):
    target[idx] = source
    return target


def mean(tensor, axis=None):
    return torch.mean(tensor, dim=axis)



# SPDX-License-Identifier: MIT

from neuralmag.common import config
from neuralmag.common.engine import N, Variable, dA
from neuralmag.field_terms.field_term import FieldTerm

__all__ = ["InterlayerExchangeField"]


def swap(m, iidx):
    result = config.backend.zeros_like(m)
    result = config.backend.assign(
        result, m[:, :, iidx[1], :], (slice(None), slice(None), iidx[0], slice(None))
    )
    result = config.backend.assign(
        result, m[:, :, iidx[0], :], (slice(None), slice(None), iidx[1], slice(None))
    )
    return result


class InterlayerExchangeField(FieldTerm):
    r"""
    Effective field contribution for interface exchange up to bbiquadratic order

    .. math::

      E = - 0.5 \int_\Gamma A \vec{m} \cdot \vec{m}_\text{other} \ds

    where :math:`\Gamma` denotes two coupled interfaces and
    :math:`\vec{m}_\text{other}` denotes the magnetization at the nearest
    point of the other interface. This expression is equivalent to the more
    common expression

    .. math::

      E = - \int_\Gamma A \vec{m}_1 \cdot \vec{m}_2 \ds

    where :math:`\Gamma` denotes a shared interface and :math:`\vec{m}_1`
    and :math:`\vec{m}_2` denote the magnetization on the respective sides.
    Currently, only surfaces in the xy-plane are supported.

    :param idx1: z-index of the first interfaces
    :type idx1: int
    :param idx2: z-index of the second interfaces
    :type idx2: int
    :param n_gauss: Degree of Gauss quadrature used in the form compiler.
    :type n_gauss: int

    :Required state attributes (if not renamed):
        * **state.material.iA** (*CCN scalar field*) Interface coupling constant in J/m^2
        * **state.material.Ms** (*cell scalar field*) The saturation magnetization in A/m
    """
    default_name = "iexchange"

    def __init__(self, idx1, idx2, **kwargs):
        super().__init__(**kwargs)
        self._iidx = [idx1, idx2]

    def register(self, state, name=None):
        super().register(state, name)

        state.iidx = config.backend.tensor(
            self._iidx, device=state.device, dtype=config.backend.integer
        )
        state.im_other = (swap, "node", (3,))

    @staticmethod
    def e_expr(m, dim):
        assert dim == 3
        iA = Variable("material__iA", "ccn")
        im_other = Variable("im_other", "nnn", (3,))

        return -0.5 * iA * m.dot(im_other) * dA(dim, idx="iidx")

    @staticmethod
    def dedm_expr(m, dim):
        assert dim == 3
        v = Variable("v", "nnn", (3,))
        iA = Variable("material__iA", "ccn")
        im_other = Variable("im_other", "nnn", (3,))

        return -iA * v.dot(im_other) * dA(dim, idx="iidx")



# SPDX-License-Identifier: MIT

import os
from time import time

import jax
import jax.numpy as jnp
import jax.numpy.fft as jfft
import numpy as np
from jax import grad, vmap
from jax.numpy import abs
from jax.numpy import arcsinh as asinh
from jax.numpy import arctan as atan
from jax.numpy import log, pi, sqrt
from scipy import constants

from neuralmag.common import logging
from neuralmag.common.engine import Variable, dV
from neuralmag.field_terms.field_term import FieldTerm

complex_dtype = {
    jnp.dtype("float32"): jnp.complex64,
    jnp.dtype("float64"): jnp.complex128,
}


def f(x, y, z):
    x, y, z = abs(x), abs(y), abs(z)
    x2, y2, z2 = x**2, y**2, z**2
    r = sqrt(x2 + y2 + z2)
    res = 1.0 / 6.0 * (2 * x2 - y2 - z2) * r
    res += jnp.nan_to_num(
        (y / 2.0 * (z2 - x2) * asinh(y / sqrt(x2 + z2))), posinf=0, neginf=0
    )
    res += jnp.nan_to_num(
        (z / 2.0 * (y2 - x2) * asinh(z / sqrt(x2 + y2))), posinf=0, neginf=0
    )
    res -= jnp.nan_to_num((x * y * z * atan(y * z / (x * r))), posinf=0, neginf=0)
    return res


def g(x, y, z):
    z = abs(z)
    x2, y2, z2 = x**2, y**2, z**2
    r = sqrt(x2 + y2 + z2)
    res = -x * y * r / 3.0
    res += jnp.nan_to_num((x * y * z * asinh(z / sqrt(x2 + y2))), posinf=0, neginf=0)
    res += jnp.nan_to_num(
        (y / 6.0 * (3.0 * z2 - y2) * asinh(x / sqrt(y2 + z2))), posinf=0, neginf=0
    )
    res += jnp.nan_to_num(
        (x / 6.0 * (3.0 * z2 - x2) * asinh(y / sqrt(x2 + z2))), posinf=0, neginf=0
    )
    res -= jnp.nan_to_num((z**3 / 6.0 * atan(x * y / (z * r))), posinf=0, neginf=0)
    res -= jnp.nan_to_num((z * y2 / 2.0 * atan(x * z / (y * r))), posinf=0, neginf=0)
    res -= jnp.nan_to_num((z * x2 / 2.0 * atan(y * z / (x * r))), posinf=0, neginf=0)
    return res


def F1(func, x, y, z, dz, dZ):
    return (
        func(x, y, z + dZ)
        - func(x, y, z)
        - func(x, y, z - dz + dZ)
        + func(x, y, z - dz)
    )


def F0(func, x, y, z, dy, dY, dz, dZ):
    return (
        F1(func, x, y + dY, z, dz, dZ)
        - F1(func, x, y, z, dz, dZ)
        - F1(func, x, y - dy + dY, z, dz, dZ)
        + F1(func, x, y - dy, z, dz, dZ)
    )


def newell(func, x, y, z, dx, dy, dz, dX, dY, dZ):
    res = (
        F0(func, x, y, z, dy, dY, dz, dZ)
        - F0(func, x - dx, y, z, dy, dY, dz, dZ)
        - F0(func, x + dX, y, z, dy, dY, dz, dZ)
        + F0(func, x - dx + dX, y, z, dy, dY, dz, dZ)
    )
    return -res / (4.0 * pi * dx * dy * dz)


def dipole_f(x, y, z, dx, dy, dz, dX, dY, dZ):
    z = z + dZ / 2.0 - dz / 2.0  # diff of cell centers for non-equidistant demag
    res = (2.0 * x**2 - y**2 - z**2) * pow(x**2 + y**2 + z**2, -5.0 / 2.0)
    res = res.at[0, 0, 0].set(0.0)
    return res * dx * dy * dz / (4.0 * pi)


def dipole_g(x, y, z, dx, dy, dz, dX, dY, dZ):
    z = z + dZ / 2.0 - dz / 2.0  # diff of cell centers for non-equidistant demag
    res = 3.0 * x * y * pow(x**2 + y**2 + z**2, -5.0 / 2.0)
    res = res.at[0, 0, 0].set(0.0)
    return res * dx * dy * dz / (4.0 * pi)


def demag_f(x, y, z, dx, dy, dz, dX, dY, dZ, p):
    res = dipole_f(x, y, z, dx, dy, dz, dX, dY, dZ)
    near = (x**2 + y**2 + z**2) / max(
        dx**2 + dy**2 + dz**2, dX**2 + dY**2 + dZ**2
    ) < p**2
    res = res.at[near].set(newell(f, x[near], y[near], z[near], dx, dy, dz, dX, dY, dZ))
    return res


def demag_g(x, y, z, dx, dy, dz, dX, dY, dZ, p):
    res = dipole_g(x, y, z, dx, dy, dz, dX, dY, dZ)
    near = (x**2 + y**2 + z**2) / max(
        dx**2 + dy**2 + dz**2, dX**2 + dY**2 + dZ**2
    ) < p**2
    res = res.at[near].set(newell(g, x[near], y[near], z[near], dx, dy, dz, dX, dY, dZ))
    return res


def h_cell(N_demag, mcell, material__Ms, rho):
    dim = [i for i in range(3) if mcell.shape[i] > 1]
    s = [mcell.shape[i] * 2 for i in dim]

    hx = jnp.zeros(N_demag[0][0].shape, dtype=complex_dtype[mcell.dtype])
    hy = jnp.zeros(N_demag[0][0].shape, dtype=complex_dtype[mcell.dtype])
    hz = jnp.zeros(N_demag[0][0].shape, dtype=complex_dtype[mcell.dtype])

    for ax in range(3):
        m_pad_fft1D = jnp.fft.rfftn(
            rho * material__Ms * mcell[:, :, :, ax], axes=dim, s=s
        )
        hx += N_demag[0][ax] * m_pad_fft1D
        hy += N_demag[1][ax] * m_pad_fft1D
        hz += N_demag[2][ax] * m_pad_fft1D

    hx = jnp.fft.irfftn(hx, axes=dim)
    hy = jnp.fft.irfftn(hy, axes=dim)
    hz = jnp.fft.irfftn(hz, axes=dim)

    return jnp.stack(
        [
            hx[: mcell.shape[0], : mcell.shape[1], : mcell.shape[2]],
            hy[: mcell.shape[0], : mcell.shape[1], : mcell.shape[2]],
            hz[: mcell.shape[0], : mcell.shape[1], : mcell.shape[2]],
        ],
        axis=3,
    )


@jax.jit
def h2d(N_demag, m, material__Ms, rho):
    mcell = (
        jnp.expand_dims(
            m[1:, 1:, :] + m[:-1, 1:, :] + m[1:, :-1, :] + m[:-1, :-1, :], -2
        )
        / 4.0
    )

    # TODO behavior for unsqueezed scalars seems OK, but not so elegant?
    Ms_hcell = jnp.expand_dims(rho * material__Ms, -1) * h_cell(
        N_demag, mcell, jnp.expand_dims(material__Ms, -1), jnp.expand_dims(rho, -1)
    ).squeeze(-2)

    h = jnp.zeros(m.shape, dtype=m.dtype)
    h = h.at[:-1, :-1].add(Ms_hcell)
    h = h.at[:-1, 1:].add(Ms_hcell)
    h = h.at[1:, :-1].add(Ms_hcell)
    h = h.at[1:, 1:].add(Ms_hcell)

    mass = jnp.zeros(h.shape[:-1], dtype=h.dtype)
    mass = mass.at[:-1, :-1].add(rho * material__Ms)
    mass = mass.at[:-1, 1:].add(rho * material__Ms)
    mass = mass.at[1:, :-1].add(rho * material__Ms)
    mass = mass.at[1:, 1:].add(rho * material__Ms)

    return h / jnp.expand_dims(mass, -1)


@jax.jit
def h3d(N_demag, m, material__Ms, rho):
    mcell = (
        +m[1:, 1:, 1:, :]
        + m[:-1, 1:, 1:, :]
        + m[1:, :-1, 1:, :]
        + m[:-1, :-1, 1:, :]
        + m[1:, 1:, :-1, :]
        + m[:-1, 1:, :-1, :]
        + m[1:, :-1, :-1, :]
        + m[:-1, :-1, :-1, :]
    ) / 8.0

    Ms_hcell = jnp.expand_dims(rho * material__Ms, -1) * h_cell(
        N_demag, mcell, material__Ms, rho
    )

    h = jnp.zeros(m.shape, dtype=m.dtype)
    h = h.at[:-1, :-1, :-1].add(Ms_hcell)
    h = h.at[:-1, :-1, 1:].add(Ms_hcell)
    h = h.at[:-1, 1:, :-1].add(Ms_hcell)
    h = h.at[:-1, 1:, 1:].add(Ms_hcell)
    h = h.at[1:, :-1, :-1].add(Ms_hcell)
    h = h.at[1:, :-1, 1:].add(Ms_hcell)
    h = h.at[1:, 1:, :-1].add(Ms_hcell)
    h = h.at[1:, 1:, 1:].add(Ms_hcell)

    mass = jnp.zeros(h.shape[:-1], dtype=h.dtype)
    mass = mass.at[:-1, :-1, :-1].add(rho * material__Ms)
    mass = mass.at[:-1, :-1, 1:].add(rho * material__Ms)
    mass = mass.at[:-1, 1:, :-1].add(rho * material__Ms)
    mass = mass.at[:-1, 1:, 1:].add(rho * material__Ms)
    mass = mass.at[1:, :-1, :-1].add(rho * material__Ms)
    mass = mass.at[1:, :-1, 1:].add(rho * material__Ms)
    mass = mass.at[1:, 1:, :-1].add(rho * material__Ms)
    mass = mass.at[1:, 1:, 1:].add(rho * material__Ms)

    return h / jnp.expand_dims(mass, -1)


def init_N_component(state, perm, func, p):
    n = state.mesh.n + tuple([1] * (3 - state.mesh.dim))
    dx = np.array(state.mesh.dx)
    dx /= dx.min()  # rescale dx to avoid NaNs when using single precision

    shape = [i * 2 if i > 1 else i for i in n]
    ij = [jfft.fftfreq(n, 1 / n, dtype=jnp.float64) for n in shape]
    ij = jnp.meshgrid(*ij, indexing="ij")
    x, y, z = [ij[ind] * dx[ind] for ind in perm]
    Lx = [n[ind] * dx[ind] for ind in perm]
    dx = [dx[ind] for ind in perm]

    # TODO enable pseudo PBCs
    # offsets = [state.arange(-state.mesh.pbc[ind], state.mesh.pbc[ind]+1) for ind in perm] # offset of pseudo PBC images
    # offsets = torch.stack(torch.meshgrid(*offsets, indexing="ij"), dim=-1).flatten(end_dim=-2)
    # Nc = state.zeros(shape)
    # for offset in offsets:
    #    Nc += func(x + offset[0]*Lx[0], y + offset[1]*Lx[1], z + offset[2]*Lx[2], *dx, *dx, self._p)
    Nc = func(x, y, z, *dx, *dx, p)

    dim = [i for i in range(3) if n[i] > 1]
    if len(dim) > 0:
        Nc = jnp.fft.rfftn(Nc, axes=dim)
    return Nc.real.clone()


def init_N(state, p):
    logging.info_green(f"[DemagField]: Set up demag tensor")

    with jax.experimental.enable_x64():
        Nxx = init_N_component(state, [0, 1, 2], demag_f, p).astype(state.dtype)
        Nxy = init_N_component(state, [0, 1, 2], demag_g, p).astype(state.dtype)
        Nxz = init_N_component(state, [0, 2, 1], demag_g, p).astype(state.dtype)
        Nyy = init_N_component(state, [1, 2, 0], demag_f, p).astype(state.dtype)
        Nyz = init_N_component(state, [1, 2, 0], demag_g, p).astype(state.dtype)
        Nzz = init_N_component(state, [2, 0, 1], demag_f, p).astype(state.dtype)

    state.N_demag = [[Nxx, Nxy, Nxz], [Nxy, Nyy, Nyz], [Nxz, Nyz, Nzz]]



# SPDX-License-Identifier: MIT

import inspect
import os
import types

import numpy as np
import pyvista as pv

from neuralmag.common import CellFunction, Function, config, logging

__all__ = ["State"]


class Material:
    def __init__(self, state):
        self._state = state

    def __getattr__(self, name):
        return getattr(self._state, "material__" + name)

    def __setattr__(self, name, value):
        # don't mess with protected attributes
        if name[0] == "_":
            super().__setattr__(name, value)
            return
        return setattr(self._state, "material__" + name, value)


class State(object):
    r"""
    This class carries all information of the spatial discretization, parameters
    and the current state of the simulation.

    :param mesh: The mesh for the simulation
    :type mesh: class:`Mesh`
    :param device: The device to be used, defaults to "cpu" (torch backend only)
    :type device: str, optional
    :param dtype: The dtype to be used, defaults to "float64" (torch backend only)
    :type dtype: str, optional
    """

    def __init__(self, mesh, device=None, dtype=None):
        self._attr_values = {}
        self._attr_types = {}
        self._attr_funcs = {}

        if device == None:
            self._device = config.device
        else:
            self._device = config.backend.device_for_state(device)

        if dtype == None:
            self._dtype = config.dtype
        else:
            self._dtype = config.backend.dtype_for_state(dtype)

        self._material = Material(self)
        self._mesh = mesh
        self.dx = self.tensor(mesh.dx)
        self.t = 0.0

        # initialize density fields for volume and facet measure with 1
        self.rho = CellFunction(self).fill(1.0, expand=True)
        if mesh.dim == 3:
            self.rhoxy = Function(self, "ccn").fill(1.0, expand=True)
            self.rhoxz = Function(self, "cnc").fill(1.0, expand=True)
            self.rhoyz = Function(self, "ncc").fill(1.0, expand=True)

        self._attr_values["eps"] = config.backend.eps(self.dtype)
        logging.info_green(
            f"[State] Running on device: {self.dx.device} (dtype = {self.dx.dtype}, backend = {config.backend_name})"
        )

    @property
    def device(self):
        """
        The PyTorch device used for all tensors.
        """
        return self._device

    @property
    def dtype(self):
        """
        The PyTorch dtype used for all tensors.
        """
        return self._dtype

    @property
    def mesh(self):
        """
        The mesh
        """
        return self._mesh

    @property
    def material(self):
        """
        The material namespace.

        The namespace supports the same functionality
        as the :class:`State` class to set and get regular and dynamic attributes.

        :Example:
            .. code-block::

                 mesh = nm.Mesh((10, 10, 1), (5e-9, 5e-9, 3e-9))
                 state = nm.State(mesh)

                 # Set saturation magnetization Ms according to Bloch's law
                 Ms0 = 8e5
                 Tc = 400.0
                 state.T = 200.0
                 state.material.Ms = lambda T: Ms0 * (1 - T/Tc)**1.5
        """
        return self._material

    def getattr(self, name):
        """
        Returns the attribute for the given name. Attributes in namespaces
        can be accessed by using "." as a seperator, e.g. :code:`material.Ms`.

        :param name: The name of the attribute
        :type name: str
        :return: The value of the attribute
        """
        container = self
        while "." in name:
            parent, child = name.split(".", 1)
            container = getattr(container, parent)
            name = child
        return getattr(container, name)

    def tensor(self, value, **kwargs):
        """
        Creates a PyTorch tensor with device and dtype set according to the
        state defaults.

        :param value: The value of the tensor
        :type value: config.backend.Tensor, list
        :return: The tensor
        :rtype: :class:`config.backend.Tensor`
        """
        default_options = {"device": self.device, "dtype": self.dtype}
        options = {**default_options, **kwargs}
        return config.backend.tensor(value, **options)

    def zeros(self, shape, **kwargs):
        """
        Creates an empty tensor of given shape with default dtype on the default device.

        :param shape: The shape of the tensor
        :type shape: tuple
        :param **kwargs: Parameters passed to the PyTorch routine
        :return: The tensor
        :rtype: config.backend.Tensor
        """
        return config.backend.zeros(
            shape, device=self.device, dtype=self.dtype, **kwargs
        )

    def __getattr__(self, name):
        if callable(self._attr_values[name]):
            if not name in self._attr_funcs:
                attr = self._attr_values[name]
                self._attr_funcs[name] = self.resolve(attr)

            func = self._attr_funcs[name]
            args = []
            for arg in list(inspect.signature(func).parameters.keys()):
                attr = getattr(self, arg)
                if hasattr(attr, "tensor"):
                    args.append(attr.tensor)
                else:
                    args.append(attr)

            value = func(*args)
        else:
            value = self._attr_values[name]

        if name in self._attr_types:
            spaces, shape = self._attr_types[name]
            return Function(self, spaces=spaces, shape=shape, tensor=value)
        else:
            return value

    def __setattr__(self, name, value):
        # don't mess with protected attributes
        if name[0] == "_":
            super().__setattr__(name, value)
            return

        if isinstance(value, (int, float)):
            value = self.tensor(value)

        if isinstance(value, tuple) and len(value) == 3:
            self._attr_types[name] = value[1:]
            value = value[0]
        else:
            self._attr_types.pop(name, None)

        if isinstance(value, list):
            try:
                value = self.tensor(value)
            except ValueError:
                pass

        if callable(value):
            self._attr_funcs.clear()

        self._attr_values[name] = value

    def _collect_func_deps(self, attr, exclude=None):
        exclude = exclude or []
        func_names = []
        args = {}
        for arg in set(inspect.signature(attr).parameters.keys()) - set(exclude):
            attr = self._attr_values[arg]

            if callable(attr):
                func_names.append(arg)
                subfunc_names, subargs = self._collect_func_deps(attr)
                func_names = [
                    f for f in func_names if f not in subfunc_names
                ] + subfunc_names
                args.update(subargs)
            else:
                args[arg] = attr

        return func_names, args

    def resolve(self, func, func_args=None):
        """
        Analyse arguments of supplied function and create Python function that
        depends solely on func_args if provided. If func_args is None the returned
        function will depend on all static state attributes that are dependencies of func.

        :param f: The function to by analyzed, if string is provided the state
                  attribute with the respective name is used.
        :type f: Callable, str
        :param func_args: Arguments of the returned function. If not set,
                          function takes all dependent attributes
        :type func_args: list, optional
        :return: New function that takes args as arguments, if args is None
                 the functions has all dependencies as arguments.

        :rtype: tuple
        """
        if isinstance(func, str):
            func = self._attr_values[func]

        subfunc_names, args = self._collect_func_deps(func, func_args)
        name = func.__name__
        name = "lmda" if func.__name__ == "<lambda>" else name

        # setup function with all dependencies
        if subfunc_names or func_args is not None:
            code = (
                f"def {name}({', '.join(args if func_args is None else func_args)}):\n"
            )
            globals = {}
            for subfunc_name in reversed(subfunc_names):
                subfunc = self._attr_values[subfunc_name]
                globals[f"__{subfunc_name}"] = subfunc
                code += (
                    f"    {subfunc_name} ="
                    f" __{subfunc_name}({', '.join(list(inspect.signature(subfunc).parameters.keys()))})\n"
                )
            globals[f"__{name}"] = func
            code += (
                "    return"
                f" __{name}({', '.join(list(inspect.signature(func).parameters.keys()))})\n"
            )

            # populate globals with bound variables
            if func_args is not None:
                for arg in list(set(args) - set(func_args)):
                    attr = getattr(self, arg)
                    if hasattr(attr, "tensor"):
                        globals[arg] = attr.tensor
                    else:
                        globals[arg] = attr

            compiled_code = compile(code, "<string>", "exec")
            return types.FunctionType(compiled_code.co_consts[0], globals, name)
        else:
            return func

    @staticmethod
    def wrap_func(f, mapping):
        """
        Wrappes a given function into another function renaming its arguments
        according to the mapping provided.

        :param f: The function to be wrapped
        :type f: Callable
        :param mapping: The name mapping of the arguments
        :type mapping: dict
        :return: The wrapped function
        :rtype: Callable

        :Example:
            .. code-block::

                state = nm.State(nm.Mesh((10, 10, 10), (1e-9, 1e-9, 1e-9)))

                def f(a, b):
                   return a + b

                g = state.wrap_func(f, {"a": "x", "b": "y"})
                # g is a function with arguments "x" and "y"
        """
        name = "lmda" if f.__name__ == "<lambda>" else f.__name__
        old_args = list(inspect.signature(f).parameters.keys())
        new_args = [mapping.get(a, a) for a in old_args]

        code = f"def {name}({', '.join(new_args)}):\n"
        code += f"    return __{name}({', '.join(new_args)})\n"

        compiled_code = compile(code, "<string>", "exec")
        return types.FunctionType(compiled_code.co_consts[0], {f"__{name}": f}, name)

    def coordinates(self, spaces=None, numpy=False):
        """
        Returns 3 tensors containing the x, y, z coordinates of each cell/node
        of the mesh. In the case of cell discretization the coordinates at the
        cell centers are provided. In the case of node discretization the node
        positions are returned.

        :param spaces: function spaces, e.g. "ccc", "nnn"
        :type spaces: str
        :param numpy: return numpy arrays instead of backend arrays
        :type numpy: bool
        :return: The coordinates
        :rtype: config.backend.Tensor

        :Example:
            .. code-block::

                state = nm.State(nm.Mesh((10, 10, 10), (1e-9, 1e-9, 1e-9)))
                x, y, z = state.coordinates('nnn')

                # initialize magnetization based on coordinate function
                state.m = VectorFunction(state)
                state.m.tensor[..., 0] = torch.sin(x/20e-9)
                state.m.tensor[..., 1] = torch.cos(x/20e-9)
        """
        if spaces == None:
            spaces = "c" * self.mesh.dim

        ranges = []
        for i, space in enumerate(spaces):
            if space == "c":
                ranges.append(
                    config.backend.linspace(
                        self.dx[i] / 2.0 + self.mesh.origin[i],
                        self.dx[i] / 2.0
                        + self.mesh.origin[i]
                        + self.dx[i] * (self.mesh.n[i] - 1.0),
                        self.mesh.n[i],
                        device=self.device,
                        dtype=self.dtype,
                    )
                )
            elif space == "n":
                ranges.append(
                    config.backend.linspace(
                        self.mesh.origin[i],
                        self.mesh.origin[i] + self.dx[i] * self.mesh.n[i],
                        self.mesh.n[i] + 1,
                        device=self.device,
                        dtype=self.dtype,
                    )
                )
            else:
                raise NotImplementedError(f"Unknown function space '{space}'.")

        coords = config.backend.meshgrid(*ranges, indexing="ij")
        if numpy:
            return tuple(config.backend.to_numpy(c) for c in coords)
        else:
            return coords

    def write_vti(self, fields, filename):
        """
        Write field data into VTI file.

        :param fields: The field data to be written. The field can be provided
            either as a :class:`Function` or as attribute name(s) of state
            attributes
        :type fields: str, Function, list
        :param filename: The name of the VTI file
        :type filename: str

        :Example:
            .. code-block::

                state = nm.State(nm.Mesh((10, 10, 10), (1e-9, 1e-9, 1e-9)))

                state.material.Ms = CellFunction(state).fill(8e5)
                state.m = VectorFunction(state).fill([0, 0, 1])

                # Write m into m.vti
                state.write_vti("m", "m.vti")

                # Write Ms and m into data.vti
                state.write_vti(["material.Ms", "m"], "data.vti")

                # Write some function f into f.vti
                f = Function(state)
                state.write_vti(f, "f.vti")
        """
        if isinstance(fields, (Function, str)):
            fields = [fields]

        n = np.array(self.mesh.n + tuple([1] * (3 - self.mesh.dim))) + 1
        grid = pv.ImageData(dimensions=n, spacing=self.mesh.dx, origin=self.mesh.origin)

        for field in fields:
            if isinstance(field, str):
                name = field
                field = self.getattr(name)
            else:
                name = field.name

            # check for spatial dimension and pure cell/node data
            if len(field.spaces) > 3:
                raise AttributeError(
                    "VTI only supports spatial dimensions smaller or equal than 3"
                )
            if len(set(field.spaces)) > 1:
                raise AttributeError(
                    "VTI only supports pure cell/nodal function spaces"
                )
            else:
                space = field.spaces[0]

            data = config.backend.to_numpy(field.tensor)

            # extend data to length 2 in hidden dimensions in case of nodal discretization
            if space == "n":
                missing_dims = tuple(
                    np.arange(3 - len(field.spaces)) + len(field.spaces)
                )
                data = np.expand_dims(data, missing_dims)
                new_shape = np.array(data.shape)
                new_shape[
                    missing_dims,
                ] = 2
                data = np.broadcast_to(data, new_shape)

            if field.shape == ():
                data = data.flatten("F")
            elif field.shape == (3,):
                data = data.reshape(-1, 3, order="F")
            else:
                raise NotImplemented(f"Unsupported shape '{field.shape}'.")

            if space == "n":
                grid.point_data.set_array(data, name)
            elif space == "c":
                grid.cell_data.set_array(data, name)
            else:
                raise NotImplemented(f"Unsupported space '{field.spaces}'.")

        dirname = os.path.dirname(filename)
        if dirname and not os.path.isdir(dirname):
            os.makedirs(dirname)

        grid.save(filename)

    def read_vti(self, filename, name=None):
        """
        Read field data from VTI file.

        :param filename: The filename of the VTI file
        :type filename: str
        :param name: The name of the attribute in the VTI file. If not
            provided, the first field in the VTI file will be read.
        :type name: str, None
        :return: The function
        :rvalue: :class:`Function`
        """
        fields = {}
        data = pv.read(filename)

        assert np.array_equal(
            self.mesh.n + (1,) * (3 - self.mesh.dim), np.array(data.dimensions) - 1
        )

        if name is None:
            name = data.array_names[0]

        n = self.mesh.n + (1,) * (3 - self.mesh.dim)
        if name in data.point_data.keys():
            spaces = "n" * self.mesh.dim
            n = tuple([x + 1 for x in n])
        elif name in data.cell_data.keys():
            spaces = "c" * self.mesh.dim
        else:
            raise RuntimeError(f"Field '{name}' not found in VTI file.")

        vals = data.get_array(name)
        if len(vals.shape) == 1:
            dim = n
            shape = ()
        else:
            dim = n + (vals.shape[-1],)
            shape = (3,)

        values = self.tensor(vals.reshape(dim, order="F"))
        if self.mesh.dim == 1:
            values = values[:, 0, 0, ...]
        if self.mesh.dim == 2:
            values = values[:, :, 0, ...]

        return Function(self, spaces=spaces, shape=shape, tensor=values)

    def domains_from_file(self, filename, scale=1.0):
        mesh = self.mesh

        # read image data and volume domains
        unstructured_mesh = pv.read(filename)

        # interpolate on mesh
        x = np.arange(mesh.n[0]) * mesh.dx[0] + mesh.dx[0] / 2.0 + mesh.origin[0]
        y = np.arange(mesh.n[1]) * mesh.dx[1] + mesh.dx[1] / 2.0 + mesh.origin[1]
        z = np.arange(mesh.n[2]) * mesh.dx[2] + mesh.dx[2] / 2.0 + mesh.origin[2]
        points = (
            np.stack(np.meshgrid(x, y, z, indexing="ij"), axis=-1).reshape(-1, 3)
            / scale
        )

        containing_cells = unstructured_mesh.find_containing_cell(points)
        data = unstructured_mesh.get_array(0)[containing_cells]
        data[
            containing_cells == -1
        ] = -1  # containing_cell == -1, if point is not included in any cell

        return Function(
            self, spaces="c" * mesh.dim, tensor=self.tensor(data.reshape(mesh.n))
        )



# SPDX-License-Identifier: MIT

from neuralmag.common.engine import N, Variable, dV
from neuralmag.field_terms.field_term import FieldTerm

__all__ = ["ExchangeField"]


class ExchangeField(FieldTerm):
    r"""
    Effective field contribution corresponding to the micromagnetic exchange energy

    .. math::

      E = \int_\Omega A \big( \nabla m_x^2 + \nabla m_y^2 + \nabla m_z^2 \big) \dx

    with the exchange constant :math:`A` given in units of :math:`\text{J/m}`.

    :param n_gauss: Degree of Gauss quadrature used in the form compiler.
    :type n_gauss: int

    :Required state attributes (if not renamed):
        * **state.material.A** (*cell scalar field*) The exchange constant in J/m
        * **state.material.Ms** (*cell scalar field*) The saturation magnetization in A/m
    """
    default_name = "exchange"

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

    @staticmethod
    def e_expr(m, dim):
        A = Variable("material__A", "c" * dim)
        return (
            A
            * (
                m.diff(N.x).dot(m.diff(N.x))
                + m.diff(N.y).dot(m.diff(N.y))
                + m.diff(N.z).dot(m.diff(N.z))
            )
            * dV(dim)
        )



# SPDX-License-Identifier: MIT

import torch
import torch.nn as nn
from torchdiffeq import odeint_adjoint as odeint

from neuralmag.common import Function, logging

__all__ = ["LLGSolverTorch"]


def llg_rhs(h, m, material__alpha):
    gamma_prime = 221276.14725379366 / (1.0 + material__alpha**2)
    return -gamma_prime * torch.linalg.cross(
        m, h
    ) - material__alpha * gamma_prime * torch.linalg.cross(m, torch.linalg.cross(m, h))


class LLGSolverTorch(nn.Module):
    """
    Time integrator using explicit adaptive time-stepping provided by the
    torchdiffeq library (https://github.com/rtqichen/torchdiffeq).

    :param state: The state used for the simulation
    :type state: :class:`State`
    :param scale_t: Internal scaling of time to improve numerical behavior
    :type scale_t: float, optional
    :param parameters: List a attribute names for the adjoint gradient computation.
                       Only required for optimization problems.
    :type parameters: list
    :param solver_options: Solver options passed to torchdiffeq
    :type solver_options: dict

    :Required state attributes:
        * **state.t** (*scalar*) The time in s
        * **state.h** (*nodal vector field*) The effective field in A/m
        * **state.m** (*nodal vector field*) The magnetization

    :Example:
        .. code-block::

            # create state with time and magnetization
            state = nm.State(nm.Mesh((10, 10, 10), (1e-9, 1e-9, 1e-9)))
            state.t = 0.0
            state.m = nm.VectorFunction(state).fill((1, 0, 0))

            # register constant Zeeman field as state.h
            nm.ExternalField(torch.Tensor((0, 0, 8e5))).register(state, "")

            # initiali LLGSolver
            llg = LLGSolver(state)

            # perform integration step
            llg.step(1e-12)

    """

    def __init__(
        self, state, scale_t=1e-9, parameters=None, max_steps=None, solver_options=None
    ):
        super().__init__()
        self._state = state
        self._scale_t = scale_t
        self._parameter_names = parameters or []
        self._solver_options = {"method": "dopri5", "atol": 1e-5, "rtol": 1e-5}
        self._solver_options.update(solver_options or {})

        # solver options
        self._max_steps = max_steps  # ignored for now

        self.reset()

    def reset(self):
        """
        Set up the function for the RHS evaluation of the LLG
        """
        logging.info_green("[LLGSolverTorch] Initialize RHS function")

        internal_args = ["t", "m"] + self._parameter_names

        self._parameter_values = []
        for parameter_name in self._parameter_names:
            value = self._state.getattr(parameter_name)
            if isinstance(value, Function):
                value = value.tensor

            parameter = torch.nn.Parameter(value)
            self.register_parameter(parameter_name, parameter)
            self._parameter_values.append(parameter)

        self._func = self._state.resolve(llg_rhs, internal_args)

    def forward(self, t, m):
        return self._scale_t * self._func(t * self._scale_t, m, *self._parameter_values)

    def relax(self, tol=2e7 * torch.pi, dt=1e-11):
        """
        Use time integration of the damping term to relax the magnetization into an
        energetic equilibrium. The convergence criterion is defined in terms of
        the maximum norm of dm/dt in rad/s.

        :param tol: The stopping criterion in rad/s, defaults to 2 pi / 100 ns
        :type tol: float
        :param dt: Interval for checking convergence
        :type dt: float
        """
        dt = 1e-11
        alpha = self._state.tensor(1.0)

        func = self._state.resolve(llg_rhs, ["t", "m", "material__alpha"])
        rhs = lambda t, m: self._scale_t * func(t * self._scale_t, m, alpha)

        logging.info_blue(
            f"[LLGSolverTorch] Relaxation started, initial energy E = {self._state.E:g} J"
        )

        t = self._state.tensor(
            [self._state.t / self._scale_t, (self._state.t + dt) / self._scale_t]
        )
        while rhs(t[0], self._state.m.tensor).norm(dim=-1).max() / self._scale_t > tol:
            logging.info_blue(
                f"[LLGSolverTorch] Relaxation step (max dm/dt = {rhs(t[0], self._state.m.tensor).norm(dim=-1).max() / self._scale_t:g}) 1/s"
            )
            m_next = odeint(
                rhs, self._state.m.tensor, t, adjoint_params=(), **self._solver_options
            )
            self._state.m.tensor[:] = m_next[-1]

        logging.info_blue(
            f"[LLGSolverTorch] Relaxation finished, final energy E = {self._state.E:g} J"
        )

    def step(self, dt):
        """
        Perform single integration step of LLG. Internally an adaptive time step is
        used.

        :param dt: The size of the time step
        :type dt: float
        """
        logging.info_blue(
            f"[LLGSolverTorch] Step: dt = {dt:g}s, t = {self._state.t:g}s"
        )
        t = self._state.tensor(
            [self._state.t / self._scale_t, (self._state.t + dt) / self._scale_t]
        )
        m_next = odeint(self, self._state.m.tensor, t, **self._solver_options)
        self._state.t.fill_(t[-1] * self._scale_t)
        self._state.m.tensor[:] = m_next[-1]

    def solve(self, t):
        """
        Solves the LLG for a list of target times. This routine is specifically
        meant to be used in the context of time-dependent optimization with
        objective functions depending on multiple mangetization snapshots.

        :param t: List of target times
        :type t: torch.Tensor
        """
        return odeint(
            self, self._state.m.tensor, t / self._scale_t, **self._solver_options
        )



# SPDX-License-Identifier: MIT

from .bulk_dmi_field import *
from .cubic_anisotropy_field import *
from .demag_field import *
from .exchange_field import *
from .external_field import *
from .field_term import *
from .interface_dmi_field import *
from .interlayer_exchange_field import *
from .total_field import *
from .uniaxial_anisotropy_field import *

__all__ = (
    field_term.__all__
    + bulk_dmi_field.__all__
    + demag_field.__all__
    + interface_dmi_field.__all__
    + interlayer_exchange_field.__all__
    + exchange_field.__all__
    + external_field.__all__
    + total_field.__all__
    + uniaxial_anisotropy_field.__all__
    + cubic_anisotropy_field.__all__
)



# SPDX-License-Identifier: MIT

import os
from collections.abc import Iterable
from functools import reduce

from ..common import Function, config, logging

__all__ = ["ScalarLogger"]


class ScalarLogger(object):
    """
    Simple logger class to log scalar values into a tab separated file.

    :param filename: The name of the log file
    :type filename: str
    :param columns: List of attribute names to be logged
    :type columns: list
    :param every: Write row to log file every nth call
    :type every: int

    :Example:
        .. code-block:: python

            # provide key strings with are available in state
            logger = ScalarLogger("log.dat", ["t","m"])

            # provide func(state) or tuple (name, func(state))
            logger = ScalarLogger("log.dat", [("t[ns]", lambda state: state.t*1e9)])

            # Actually log a row
            state = State(mesh)
            logger.log(state)
    """

    def __init__(self, filename, columns, every=1):
        # create directory if not existent
        if not os.path.dirname(filename) == "" and not os.path.exists(
            os.path.dirname(filename)
        ):
            try:
                os.makedirs(os.path.dirname(filename))
            except OSError as exc:  # Guard against race condition
                if exc.errno != errno.EEXIST:
                    raise

        self._filename = filename
        self._file = None
        self._every = every
        self._i = 0
        self._columns = columns

    def add_column(self, column):
        """
        Add column to log file.

        This method is automatically called for all
        columns on initialization of the logger.
        The column can be provided either as attribute name or as a Callable
        that takes the state as the only argument.
        If the column is a :class:`Function`, the functional is averaged over
        the whole mesh before logging.

        .. note::
          This method can only be called before the first line is logged

        :param column: The column to be logged
        :type column: str, Callable
        """
        if self._file is not None:
            raise RuntimeError(
                "You cannot add columns after first log row has been written."
            )
        self._columns.append(column)

    def log(self, state):
        """
        Log simulation step

        :param state: The state to be logged
        :type state: :class:`State`
        """
        self._i += 1
        if (self._i - 1) % self._every > 0:
            return

        values = []

        for column in self._columns:
            if isinstance(column, str):
                name = column
                raw_value = state.getattr(column)
            elif hasattr(column, "__call__"):
                name = column.__name__
                raw_value = column(state)
            elif isinstance(column, tuple) or isinstance(column, list):
                name = column[0]
                raw_value = column[1](state)
            else:
                raise RuntimeError("Column type not supported.")

            if isinstance(raw_value, Function):
                value = raw_value.avg().tolist()
            elif isinstance(raw_value, config.backend.Tensor):
                value = config.backend.to_numpy(raw_value).tolist()
            else:
                value = raw_value
            values.append((name, value))

        if self._file is None:
            self._file = open(self._filename, "w")
            self._write_header(values)

        self._write_row(values)

    def _write_header(self, columns):
        headings = []

        for column in columns:
            if isinstance(column[1], Iterable):
                if len(column[1]) == 3:
                    for i in ("x", "y", "z"):
                        headings.append(column[0] + "_" + i)
                else:
                    for i in range(len(column[1])):
                        headings.append(column[0] + "_" + str(i))
            else:
                headings.append(column[0])

        format_str = "#" + "    ".join(["%-22s"] * len(headings)) + "\n"
        self._file.write(format_str % tuple(headings))
        self._file.flush()

    def _write_row(self, columns):
        flat_values = reduce(
            lambda x, y: x + y,
            map(
                lambda x: tuple(x[1]) if isinstance(x[1], Iterable) else (x[1],),
                columns,
            ),
        )
        format_str = "    ".join(["%+.15e"] * len(flat_values)) + "\n"
        self._file.write(format_str % flat_values)
        self._file.flush()

    def __del__(self):
        if self._file is not None:
            self._file.close()

    def resumable_step(self):
        """
        Returns the first step that can be written when resuming, e.g. if the
        logger logs every 10th step and the first (i = 0) step was already
        logged, the result is 10.

        :return: The step number
        :rtype: int
        """
        if self._file is not None:
            raise RuntimeError(
                "Cannot resume from log file that is already open for writing."
            )

        i = 0
        with open(self._filename, "r") as f:
            for i, l in enumerate(f):
                pass
        return i * self._every

    def resume(self, i):
        """
        Try to resume existing log file from log step i. The log file
        is truncated accordingly.

        :param i: The log step to resume from
        :type i: int
        """
        number = (self.resumable_step() - i) / self._every

        # from https://superuser.com/questions/127786/efficiently-remove-the-last-two-lines-of-an-extremely-large-text-file
        count = 0
        # with open(self._filename, 'r+b') as f:
        #    f.seek(0, os.SEEK_END)
        #    end = f.tell()
        #    while f.tell() > 0:
        #        f.seek(-1, os.SEEK_CUR)
        #        char = f.read(1)
        #        if char != '\n' and f.tell() == end:
        #            raise RuntimeError("Cannot resume: logfile does not end with a newline.")
        #        if char == '\n':
        #            count += 1
        #        if count == number + 1:
        #            f.truncate()
        #            break
        #        f.seek(-1, os.SEEK_CUR)

        with open(self._filename, "r+b", buffering=0) as f:
            f.seek(0, os.SEEK_END)
            end = f.tell()
            while f.tell() > 0:
                f.seek(-1, os.SEEK_CUR)
                # print(f.tell())
                char = f.read(1)
                if char != b"\n" and f.tell() == end:
                    raise RuntimeError(
                        "Cannot resume: logfile does not end with a newline."
                    )
                    # print ("No change: file does not end with a newline")
                    # exit(1)
                if char == b"\n":
                    count += 1
                if count == number + 1:
                    f.truncate()
                    break
                    # print ("Removed " + str(number) + " lines from end of file")
                    # exit(0)
                f.seek(-1, os.SEEK_CUR)

        self._i = i
        if self._i > 0:
            self._file = open(self._filename, "a")



# SPDX-License-Identifier: MIT

import inspect
import types

from scipy import constants

from neuralmag.common import VectorFunction, config
from neuralmag.common.engine import Variable, dV
from neuralmag.field_terms.field_term import FieldTerm

__all__ = ["ExternalField"]


class ExternalField(FieldTerm):
    r"""
    Effective field contribution corresponding to the external field

    .. math::

      E = - \int_\Omega \mu_0 M_s  \vec{m} \cdot \vec{h} \dx

    :param h: The field either given as a :code:`config.backend.Tensor` or callable
              in case of a field that depends e.g. on :code:`state.t`.
              The shape must be either (nx, ny, nz, 3) or (3,) in which case
              the field expanded to full size.
    :type h: config.backend.Tensor
    :param n_gauss: Degree of Gauss quadrature used in the form compiler.
    :type n_gauss: int

    :Required state attributes (if not renamed):
        * **state.material.Ms** (*cell scalar field*) The saturation magnetization in A/m

    :Example:
        .. code-block::

            state = nm.State(nm.Mesh((10, 10, 10), (1e-9, 1e-9, 1e-9)))

            # define constant external field from expanded function
            h_ext = nm.VectorFunction(state).fill((0, 0, 0), expand=True)
            external = nm.ExternalField(h_ext)

            # define external field in y-direction linearly increasing with time
            external = nm.ExternalField(lambda t: t * state.tensor([0, 8e5 / 10e-9, 0]))

    """
    default_name = "external"
    h = None

    def __init__(self, h, **kwargs):
        super().__init__(**kwargs)
        self._h = h

    def register(self, state, name=None):
        tensor_shape = VectorFunction(state).tensor_shape
        if callable(self._h):
            # check if h function return 3-vector
            value = state.resolve(self._h, [])()
            if value.shape == (3,):
                arg_names = list(inspect.signature(self._h).parameters.keys())
                block = config.backend.CodeBlock(plain=True)
                with block.add_function("h", arg_names) as func:
                    func.retrn_expanded(f"__h({', '.join(arg_names)})", tensor_shape)
                compiled_code = compile(str(block), "<string>", "exec")
                self.h = types.FunctionType(
                    compiled_code.co_consts[0],
                    {**config.backend.libs, **{f"__h": self._h}},
                    name,
                )
            else:
                self.h = self._h
        elif isinstance(self._h, config.backend.Tensor):
            if self._h.shape == tensor_shape:
                self.h = self._h
            elif self._h.shape == (3,):
                self.h = config.backend.broadcast_to(self._h, tensor_shape)
            else:
                raise Exception("Shape not matching")
        elif isinstance(self._h, VectorFunction):
            self.h = self._h.tensor
        else:
            raise Exception("Type not supported")

        super().register(state, name)
        # fix reference to h_external in E_external if suffix is changed
        if name is not None:
            wrapped = state.wrap_func(self.E, {"h_external": self.attr_name("h", name)})
            setattr(state, self.attr_name("E", name), wrapped)

    @staticmethod
    def e_expr(m, dim):
        Ms = Variable("material__Ms", "c" * dim)
        h_external = Variable("h_external", "n" * dim, (3,))
        return -constants.mu_0 * Ms * m.dot(h_external) * dV(dim)



# SPDX-License-Identifier: MIT

import inspect
import sys

from scipy import constants

from neuralmag.common import CodeClass, Function, VectorFunction, config
from neuralmag.common import engine as en
from neuralmag.common import logging

__all__ = ["FieldTerm"]


class FieldTerm(CodeClass):
    r"""
    Base class of all effective field contributions. In simple cases,
    a subclass is just required to implement the energy functional of a field
    contribution and a default_name which is used when registering the field
    term with a state. The form compiler of NeuralMag is then used to generate
    efficient code for the computation of both the energy and effective field
    by means of a just-in-time compiler.

    :param n_gauss: Degree of Gauss quadrature used in the form compiler.
    :type n_gauss: int

    :Example:
        .. code-block::

            import neuralmag as nm
            from neuralmag.generators import pytorch_generator as gen

            # Example subclass implementing a uniaxial anisotropy
            class UniaxialAnisotropyField(nm.FieldTerm):
                default_name = "uaniso"

                @staticmethod
                def e_expr(m, dim):
                    K = en.Variable("material__Ku", "c" * dim)
                    axis = en.Variable("material__Ku_axis", "c" * dim, (3,))
                    return -K * m.dot(axis) ** 2 * en.dV(dim)

            # Use instance of class to register dynamic attributes in state
            state = nm.State(nm.Mesh((10, 10, 10), (1e-9, 1e-9, 1e-9)))
            UniaxialAnisotropyField().register(state)

            # compute field and energy
            h = state.h_uaniso
            E = state.E_uaniso
    """

    default_name = None

    def __init__(self, n_gauss=None):
        self._n_gauss = n_gauss or config.fem["n_gauss"]

    def __init_subclass__(cls, **kwargs):
        if getattr(cls, "default_name") is None:
            raise TypeError(
                f"Can't instantiate abstract class {cls.__name__} without 'default_name' attribute defined"
            )
        return super().__init_subclass__(**kwargs)

    def register(self, state, name=None):
        r"""
        Registers dynamic attributes for the computation of the effective field
        and energy with the given :class:`State` object. By naming convention,
        these methods are registered as :code:`state.h_{name}` and
        :code:`state.E_{name}` or :code:`state.h` and :code:`state.E` in case
        that of :code:`name` being an empty string.

        :param state: The state
        :type state: :class:`State`
        :param name: The name used for the registration, falls back to :code:`default_name`
                     attribute of the class.
        :type name: str, optional
        """
        dim = state.mesh.dim
        if hasattr(self, "e_expr"):
            self.save_and_load_code(self._n_gauss, dim)
        if not hasattr(self, "h"):
            self.h = config.backend.compile(self._code.h)
        if not hasattr(self, "E"):
            self.E = config.backend.compile(self._code.E)
        logging.info_green(
            f"[{self.__class__.__name__}] Register state methods (field:"
            f" '{self.attr_name('h', name)}', energy: '{self.attr_name('E', name)}')"
        )
        setattr(state, self.attr_name("h", name), (self.h, "n" * dim, (3,)))
        setattr(state, self.attr_name("E", name), self.E)

    @classmethod
    def attr_name(cls, attr, name=None):
        r"""
        Returns the attribute name for a given attribute, using the :class:`default_name` attribute
        of the class.

        :param attr: The attribute name, e.g. "E" or "h"
        :type attr: str
        :param name: The name of the class, defaults to :class:`cls.default_name`
        :type attr: str
        """
        if name is None:
            name = cls.default_name
        if name == "":
            return attr
        return f"{attr}_{name}"

    @classmethod
    def _generate_code(cls, n_gauss, dim):
        code = config.backend.CodeBlock()
        m = en.Variable("m", "n" * dim, (3,))

        if not hasattr(cls, "h"):
            # generate linear-form cmds
            if hasattr(cls, "dedm_expr"):
                field_expr = cls.dedm_expr(m, dim)
            else:
                field_expr = en.gateaux_derivative(cls.e_expr(m, dim), m)

            cmds1, vars1 = en.linear_form_cmds(field_expr, n_gauss)

            # generate lumped mass cmds
            v = en.Variable("v", "n" * dim)
            Ms = en.Variable("material__Ms", "c" * dim)
            cmds2, vars2 = en.linear_form_cmds(-constants.mu_0 * Ms * v * en.dV(dim))

            with code.add_function("h", sorted(list(vars1 | vars2 | {"m"}))) as f:
                f.zeros_like("h", "m")
                for cmd in cmds1:
                    f.add_to("h", cmd[0], cmd[1])

                f.zeros_like("mass", "m", shape="m.shape[:-1]")
                for cmd in cmds2:
                    f.add_to("mass", cmd[0], cmd[1])

                f.retrn("h / mass.reshape(mass.shape + (1,))")  # unsqueeze(-1)")

        if not hasattr(cls, "E"):
            terms, variables = en.compile_functional(cls.e_expr(m, dim), n_gauss)
            with code.add_function("E", variables) as f:
                f.retrn_sum(*[term["cmd"] for term in terms])

        return code



# SPDX-License-Identifier: MIT

import equinox as eqx
import jax
import jax.numpy as jnp
from diffrax import Dopri5, Event, ODETerm, PIDController, SaveAt, diffeqsolve

from neuralmag.common import Function, logging

__all__ = ["LLGSolverJAX"]


def llg_rhs(h, m, material__alpha):
    gamma_prime = 221276.14725379366 / (1.0 + material__alpha**2)
    return -gamma_prime * jnp.cross(m, h) - material__alpha * gamma_prime * jnp.cross(
        m, jnp.cross(m, h)
    )


class LLGSolverJAX(object):
    """
    Time integrator using explicit adaptive time-stepping provided by the
    torchdiffeq library (https://github.com/rtqichen/torchdiffeq).

    :param state: The state used for the simulation
    :type state: :class:`State`
    :param scale_t: Internal scaling of time to improve numerical behavior
    :type scale_t: float, optional
    :param parameters: List a attribute names for the adjoint gradient computation.
                       Only required for optimization problems.
    :type parameters: list

    :Required state attributes:
        * **state.t** (*scalar*) The time in s
        * **state.h** (*nodal vector field*) The effective field in A/m
        * **state.m** (*nodal vector field*) The magnetization
    """

    def __init__(self, state, scale_t=1e-9, parameters=None, max_steps=4096):
        super().__init__()
        self._state = state
        self._scale_t = scale_t
        self._parameters = [] if parameters is None else parameters
        self._dt0 = 1e-14

        # solver options
        self._max_steps = max_steps

        # TODO Solver options
        # self._solver_options = {"method": "dopri5", "atol": 1e-5, "rtol": 1e-5}
        self._solver = Dopri5()
        self._stepsize_controller = PIDController(rtol=1e-5, atol=1e-5)
        self._saveat_step = SaveAt(t1=True)

        self.reset()

    def reset(self):
        """
        Set up the function for the RHS evaluation of the LLG
        """
        logging.info_green("[LLGSolverJAX] Initialize RHS function")

        internal_args = ["t", "m"] + self._parameters

        self._func = self._state.resolve(llg_rhs, internal_args)
        rhs = lambda t, m, args: self._scale_t * self._func(t * self._scale_t, m, *args)
        self._term = ODETerm(jax.jit(rhs))
        self._solver_state = None

    def relax(self, tol=2e7 * jnp.pi, dt=1e-11):
        """
        Use time integration of the damping term to relax the magnetization into an
        energetic equilibrium. The convergence criterion is defined in terms of
        the maximum norm of dm/dt in rad/s.

        :param tol: The stopping criterion in rad/s, defaults to 2 pi / 100 ns
        :type tol: float
        :param dt: Interval for checking convergence
        :type dt: float
        """
        alpha = self._state.tensor(1.0)

        func = self._state.resolve(llg_rhs, ["t", "m", "material__alpha"])
        rhs = lambda t, m, _: self._scale_t * func(t * self._scale_t, m, alpha)
        term = ODETerm(jax.jit(rhs))

        logging.info_blue(
            f"[LLGSolverJAX] Relaxation started, initial energy E = {self._state.E:g} J"
        )
        t = self._scale_t * self._state.t
        # rhs_args = [self._scale_t * self._state.t, self._state.m.tensor, None]
        while (
            jnp.linalg.norm(
                term.vector_field(t, self._state.m.tensor, None), axis=-1
            ).max()
            / self._scale_t
            > tol
        ):
            logging.info_blue(
                f"[LLGSolverJAX] Relaxation step (max dm/dt = {jnp.linalg.norm(term.vector_field(t, self._state.m.tensor, None), axis=-1).max() / self._scale_t:g}) 1/s"
            )
            sol = diffeqsolve(
                term,
                self._solver,
                t0=self._state.t / self._scale_t,
                t1=(self._state.t + dt) / self._scale_t,
                dt0=self._dt0 / self._scale_t,
                y0=self._state.m.tensor,
                saveat=self._saveat_step,
                stepsize_controller=self._stepsize_controller,
                max_steps=self._max_steps,
            )
            self._state.m.tensor = sol.ys[-1]

        logging.info_blue(
            f"[LLGSolverJAX] Relaxation finished, final energy E = {self._state.E:g} J"
        )

    def step(self, dt, *args):
        """
        Perform single integration step of LLG. Internally an adaptive time step is
        used.

        :param dt: The size of the time step
        :type dt: float
        TODO args
        """
        logging.info_blue(f"[LLGSolverJAX] Step: dt = {dt:g}s, t = {self._state.t:g}s")

        sol = diffeqsolve(
            self._term,
            self._solver,
            t0=self._state.t / self._scale_t,
            t1=(self._state.t + dt) / self._scale_t,
            dt0=self._dt0 / self._scale_t,
            y0=self._state.m.tensor,
            args=args,
            saveat=self._saveat_step,
            stepsize_controller=self._stepsize_controller,
            solver_state=self._solver_state,
            max_steps=self._max_steps,
        )
        self._solver_state = sol.solver_state
        self._state.t = sol.ts[-1] * self._scale_t
        self._state.m.tensor = sol.ys[-1]
        return sol

    def solve(self, t, *args):
        """
        Solves the LLG for a list of target times. This routine is specifically
        meant to be used in the context of time-dependent optimization with
        objective functions depending on multiple mangetization snapshots.

        :param t: List of target times
        :type t: torch.Tensor
        TODO args
        """
        t_scaled = t / self._scale_t
        saveat = SaveAt(ts=t_scaled)
        sol = diffeqsolve(
            self._term,
            self._solver,
            t0=t_scaled[0],
            t1=t_scaled[-1],
            dt0=self._dt0 / self._scale_t,
            y0=self._state.m.tensor,
            args=args,
            saveat=saveat,
            stepsize_controller=self._stepsize_controller,
            max_steps=self._max_steps,
        )
        return sol



# SPDX-License-Identifier: MIT

from neuralmag.common.engine import N, Variable, dV
from neuralmag.field_terms.field_term import FieldTerm

__all__ = ["UniaxialAnisotropyField"]


class UniaxialAnisotropyField(FieldTerm):
    r"""
    Effective field contribution corresponding to the quadratic uniaxial anisotropy energy

    .. math::
      E = - \int_\Omega K \big( \vec{m} \cdot \vec{e}_k \big)^2 \dx

    with the anisotropy constant :math:`K` given in units of :math:`\text{J/m}^3`.
    For higher order anisotropy, use the :class:`UniaxialAnisotropyField2`.

    :param n_gauss: Degree of Gauss quadrature used in the form compiler.
    :type n_gauss: int

    :Required state attributes (if not renamed):
        * **state.material.Ku** (*cell scalar field*) The anisotropy constant in J/m^3
        * **state.material.Ku_axis** (*cell vector field*) The anisotropy axis as unit vector field
        * **state.material.Ms** (*cell scalar field*) The saturation magnetization in A/m
    """
    default_name = "uaniso"

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

    @staticmethod
    def e_expr(m, dim):
        K = Variable("material__Ku", "c" * dim)
        axis = Variable("material__Ku_axis", "c" * dim, (3,))
        return -K * m.dot(axis) ** 2 * dV(dim)
